'''
Module with functions for making plots.
'''

import os
import sys

import corner
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from species.data import database
from species.plot import util


mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
mpl.rcParams['font.family'] = 'serif'

plt.rc('axes', edgecolor='black', linewidth=2)


def plot_walkers(tag,
                 output,
                 nsteps=None,
                 offset=None):
    '''
    :return: None
    '''

    sys.stdout.write('Plotting walkers: '+output+'...')
    sys.stdout.flush()

    species_db = database.Database()
    box = species_db.get_samples(tag)

    samples = box.samples
    labels = util.update_labels(box.parameters)

    ndim = samples.shape[-1]

    plt.figure(1, figsize=(6, 5))
    gridsp = mpl.gridspec.GridSpec(4, 1)
    gridsp.update(wspace=0, hspace=0.1, left=0, right=1, bottom=0, top=1)

    for i in range(ndim):
        ax = plt.subplot(gridsp[i, 0])

        ax.grid(True, linestyle=':', linewidth=0.7, color='silver', dashes=(1, 4))

        if i == ndim-1:
            ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=5, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=True)

            ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=3, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=True)

        else:
            ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=5, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=False)

            ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=3, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=False)

        if i == ndim-1:
            ax.set_xlabel('Step number', fontsize=10)
        else:
            ax.set_xlabel('', fontsize=10)

        ax.set_ylabel(labels[i], fontsize=10)

        if offset:
            ax.get_xaxis().set_label_coords(0.5, offset[0])
            ax.get_yaxis().set_label_coords(offset[1], 0.5)
        else:
            ax.get_xaxis().set_label_coords(0.5, -0.22)
            ax.get_yaxis().set_label_coords(-0.09, 0.5)

        if nsteps:
            ax.set_xlim(0, nsteps)

        for j in range(samples.shape[0]):
            ax.plot(samples[j, :, i], ls='-', lw=0.5, color='black', alpha=0.5)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.close()

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()


def plot_posterior(tag,
                   burnin,
                   output,
                   title=None,
                   offset=None,
                   title_fmt='.2f'):
    '''
    :return: None
    '''

    sys.stdout.write('Plotting posteriors: '+output+'...')
    sys.stdout.flush()

    species_db = database.Database()
    box = species_db.get_samples(tag)

    samples = box.samples
    labels = util.update_labels(box.parameters)

    ndim = samples.shape[-1]

    samples = samples[:, int(burnin):, :].reshape((-1, ndim))

    fig = corner.corner(samples, labels=labels, quantiles=[0.16, 0.5, 0.84],
                        label_kwargs={'fontsize': 13}, show_titles=True,
                        title_kwargs={'fontsize': 12}, title_fmt=title_fmt)

    axes = np.array(fig.axes).reshape((ndim, ndim))

    for i in range(ndim):
        for j in range(ndim):
            ax = axes[i, j]

            ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=5, labelsize=12, top=True,
                           bottom=True, left=True, right=True)

            ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                           direction='in', width=0.8, length=3, labelsize=12, top=True,
                           bottom=True, left=True, right=True)

        if offset:
            ax.get_xaxis().set_label_coords(0.5, offset[0])
            ax.get_yaxis().set_label_coords(offset[1], 0.5)
        else:
            ax.get_xaxis().set_label_coords(0.5, -0.26)
            ax.get_yaxis().set_label_coords(-0.27, 0.5)


    par_val = box.chisquare

    for i in range(ndim):
        ax = axes[i, i]
        ax.axvline(par_val[i], color='tomato')

        for j in range(i+1, ndim):
            ax = axes[j, i]
            ax.axvline(par_val[i], color='tomato')
            ax.axhline(par_val[j], color='tomato')
            ax.plot(par_val[i], par_val[j], 's', color='tomato')

    if title:
        fig.suptitle(title, y=1.02, fontsize=16)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.close()

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()



'''
Text
'''

import os
import sys
import time

import numpy as np

from numpy import ma
from astropy import units as u
from astropy.coordinates import SkyCoord

from astroquery.simbad import Simbad
from astroquery.vizier import Vizier


class NoStdStreams:
    '''
    Text
    '''

    def __init__(self, stdout=None, stderr=None):
        self.devnull = open(os.devnull, 'w')
        self._stdout = stdout or self.devnull or sys.stdout
        self._stderr = stderr or self.devnull or sys.stderr
        self.old_stdout = None
        self.old_stderr = None

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.devnull.close()

with NoStdStreams():
    from astroquery.gaia import Gaia


def get_simbad(name):
    '''
    Function for getting the Simbad identifier of an object.

    :param name:
    :type name: numpy.ndarray

    :return: Simbad name.
    :rtype: numpy.ndarray
    '''

    simbad_id = []

    for item in name:
        time.sleep(0.15)

        simbad = Simbad.query_object(item)

        if simbad is None:
            sim_id = None

        else:
            sim_id = simbad['MAIN_ID'][0]
            # sim_id = sim_id.decode('utf-8')

        simbad_id.append(sim_id)

    return np.asarray(simbad_id, dtype=np.str_)

def get_distance(target):
    '''
    :param target: Target name.
    :type target: str

    :return: SIMBAD name and distance (pc).
    :rtype: str, float
    '''

    # Liu et al. (2016)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/833/96

    # Dupuy & Liu (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJS/201/19

    # Faherty et al. (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/752/56

    # Weinberger et al. (2016)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/AJ/152/24

    # Theissen (2018)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/862/173

    # Kirkpatrick et al. (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/753/156

    # Monet et al. (1992)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/AJ/103/638

    catalogues = ('J/ApJ/833/96'
                  'J/ApJS/201/19',
                  'J/ApJ/752/56',
                  'J/AJ/152/24',
                  'J/ApJ/862/173',
                  'J/ApJ/753/156',
                  'J/AJ/103/638')

    if target[-2:] == 'AB':
        target = target[:-2]

    elif target[-3:] == 'ABC':
        target = target[:-3]

    if target[0:7] == 'DENIS-P':
        target = target[:5]+target[7:]

        if target[-2] == '.':
            target = target[:-2]

    if target[0:5] == 'DENIS' and target[6:7] != 'J':
        target = target[:5]+' J'+target[6:]

    simbad_id = 'None'
    parallax = ma.masked

    time.sleep(0.15)

    simbad = Simbad()
    simbad.add_votable_fields('parallax')

    simbad = simbad.query_object(target)

    if simbad is not None:
        simbad_id = simbad['MAIN_ID'][0]
        simbad_id = simbad_id.decode('utf-8')
        parallax = simbad['PLX_VALUE'][0]

    if ma.is_masked(parallax):
        for _, item in enumerate(catalogues):
            result = Vizier.query_object(target, catalog=item)

            if result.keys():
                try:
                    parallax = result[0]['plx'][0] # [mas]
                except KeyError:
                    pass

                if ma.is_masked(parallax):
                    try:
                        parallax = result[0]['Plx'][0] # [mas]
                    except KeyError:
                        pass

            else:
                continue

            if not ma.is_masked(parallax):
                break

    if ma.is_masked(parallax):

        if simbad is not None:
            coord_ra = simbad['RA'][0]
            coord_dec = simbad['DEC'][0]

            coord = SkyCoord(ra=coord_ra, dec=coord_dec, unit=(u.hourangle, u.deg), frame='icrs')
            result = Gaia.query_object(coordinate=coord, width=1.*u.arcsec, height=1.*u.arcsec)

            if result:
                parallax = result['parallax'][0] # [mas]

    if ma.is_masked(parallax) or parallax < 0.:
        distance = np.nan
    else:
        distance = 1./(parallax*1e-3) # [pc]

    return simbad_id, distance



'''
Text
'''

import os
import sys

from urllib.request import urlretrieve

import h5py
import numpy as np

from astropy.io import fits

from species.data import queries, util


def add_vlm_plx(input_path,
                database):
    '''
    Function for adding the Database of Ultracool Parallaxes to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    data_file = os.path.join(input_path, 'vlm-plx-all.fits')

    url = 'http://www.as.utexas.edu/~tdupuy/plx/' \
          'Database_of_Ultracool_Parallaxes_files/vlm-plx-all.fits'

    if not os.path.isfile(data_file):
        sys.stdout.write('Downloading Database of Ultracool Parallaxes (307 kB)...')
        sys.stdout.flush()

        urlretrieve(url, data_file)

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    sys.stdout.write('Adding Database of Ultracool Parallaxes...')
    sys.stdout.flush()

    group = 'photometry/vlm-plx'

    database.create_group(group)

    hdulist = fits.open(data_file)
    photdata = hdulist[1].data

    plx = photdata['PLX'] # [mas]
    distance = 1./(plx*1e-3) # [pc]

    name = photdata['NAME']
    name = np.core.defchararray.strip(name)

    sptype = photdata['ISPTSTR']
    sptype = np.core.defchararray.strip(sptype)

    sptype_op = photdata['OSPTSTR']
    sptype_op = np.core.defchararray.strip(sptype_op)

    for i, item in enumerate(sptype):
        if item == 'null':
            sptype[i] = sptype_op[i]

    flag = photdata['FLAG']
    flag = np.core.defchararray.strip(flag)

    sptype = util.update_sptype(sptype)

    dtype = h5py.special_dtype(vlen=bytes)

    dset = database.create_dataset(group+'/name', (np.size(name), ), dtype=dtype)
    dset[...] = name

    dset = database.create_dataset(group+'/sptype', (np.size(sptype), ), dtype=dtype)
    dset[...] = sptype

    dset = database.create_dataset(group+'/flag', (np.size(flag), ), dtype=dtype)
    dset[...] = flag

    database.create_dataset(group+'/distance', data=distance, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Y', data=photdata['YMAG'], dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.J', data=photdata['JMAG'], dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.H', data=photdata['HMAG'], dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.K', data=photdata['KMAG'], dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Lp', data=photdata['LMAG'], dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Mp', data=photdata['MMAG'], dtype='f')
    database.create_dataset(group+'/2MASS/2MASS.J', data=photdata['J2MAG'], dtype='f')
    database.create_dataset(group+'/2MASS/2MASS.H', data=photdata['H2MAG'], dtype='f')
    database.create_dataset(group+'/2MASS/2MASS.Ks', data=photdata['K2MAG'], dtype='f')

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    sys.stdout.write('Querying SIMBAD...')
    sys.stdout.flush()

    simbad_id = queries.get_simbad(name)

    dset = database.create_dataset(group+'/simbad', (np.size(simbad_id), ), dtype=dtype)
    dset[...] = simbad_id

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    database.close()



'''
Module for DRIFT-PHOENIX atmospheric models.
'''

import os
import sys
import tarfile

from urllib.request import urlretrieve

import numpy as np

from species.data import util


def add_drift_phoenix(input_path,
                      database):
    '''
    Function for adding the DRIFT-PHOENIX atmospheric models to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_file = os.path.join(input_path, 'drift-phoenix.tgz')
    data_folder = os.path.join(input_path, 'drift-phoenix/')

    url = 'https://people.phys.ethz.ch/~stolkert/species/drift-phoenix.tgz'

    if not os.path.isfile(data_file):
        sys.stdout.write('Downloading DRIFT-PHOENIX model spectra (151 MB)...')
        sys.stdout.flush()

        urlretrieve(url, data_file)

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    sys.stdout.write('Unpacking DRIFT-PHOENIX model spectra...')
    sys.stdout.flush()

    tar = tarfile.open(data_file)
    tar.extractall(input_path)
    tar.close()

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    teff = []
    logg = []
    feh = []
    wavelength = None
    flux = []

    for _, _, file_list in os.walk(data_folder):
        for filename in sorted(file_list):

            if filename.startswith('lte_'):
                sys.stdout.write('\rAdding DRIFT-PHOENIX model spectra... '+filename)
                sys.stdout.flush()

                teff.append(float(filename[4:8]))
                logg.append(float(filename[9:12]))
                feh.append(float(filename[12:16]))

                data = np.loadtxt(data_folder+filename)

                if wavelength is None:
                    # [Angstrom] -> [micron]
                    wavelength = data[:, 0]*1e-4

                # [erg s-1 cm-2 Angstrom-1] -> [W m-2 micron-1]
                flux.append(data[:, 1]*1e-7*1e4*1e4)

    data_sorted = util.sort_data(np.asarray(teff),
                                 np.asarray(logg),
                                 np.asarray(feh),
                                 wavelength,
                                 np.asarray(flux))

    util.write_data('drift-phoenix', database, data_sorted)

    sys.stdout.write('\rAdding DRIFT-PHOENIX model spectra... [DONE]                    \n')
    sys.stdout.flush()



'''
Read module.
'''

import os
import configparser

import h5py
import numpy as np

from scipy.interpolate import interp1d

from species.analysis import photometry
from species.core import box
from species.read import read_filter


class ReadCalibration:
    '''
    Text
    '''

    def __init__(self,
                 spectrum,
                 filter_name):
        '''
        :param spectrum: Database tag of the calibration spectrum.
        :type spectrum: str
        :param filter_name: Filter ID. Full spectrum is used if set to None.
        :type filter_name: str

        :return: None
        '''

        self.spectrum = spectrum
        self.filter_name = filter_name

        if filter_name is None:
            self.wl_range = None

        else:
            transmission = read_filter.ReadFilter(filter_name)
            self.wl_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def interpolate(self):
        '''
        :return: Interpolated spectrum.
        :rtype: scipy.interpolate.interpolate.interp1d
        '''

        calibbox = self.get_spectrum()

        return interp1d(calibbox.wavelength,
                        calibbox.flux,
                        kind='cubic',
                        bounds_error=False,
                        fill_value=float('nan'))

    def get_spectrum(self,
                     model_par=None,
                     negative=False):
        '''
        :param model_par: Model parameter values. Not used if set to None.
        :type model_par: dict
        :param negative: Include negative values.
        :type negative: bool

        :return: Spectrum data.
        :rtype: numpy.ndarray
        '''

        h5_file = h5py.File(self.database, 'r')

        data = h5_file['spectra/calibration/'+self.spectrum]
        data = np.asarray(data)

        if not negative:
            indices = np.where(data[1, ] > 0.)[0]
            data = data[:, indices]

        h5_file.close()

        if model_par:
            data[1, ] = model_par['offset'] + model_par['scaling']*data[1, ]

        return box.create_box(boxtype='spectrum',
                              spectrum='calibration',
                              wavelength=data[0, ],
                              flux=data[1, ],
                              name=self.spectrum,
                              simbad=None,
                              sptype=None,
                              distance=None)

    def get_photometry(self,
                       model_par,
                       synphot=None):
        '''
        :param model_par: Model parameter values.
        :type model_par: dict

        :return: Average flux density (W m-2 micron-1).
        :rtype: float
        '''

        specbox = self.get_spectrum(model_par)
        synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_photometry(specbox.wavelength, specbox.flux)



'''
Text
'''

import os
import sys

from urllib.request import urlretrieve

import h5py
import numpy as np
import pandas as pd

from species.data import util


def add_leggett(input_path,
                database):
    '''
    Function for adding the Database of Ultracool Parallaxes to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    data_file1 = os.path.join(input_path, '2010_phot.xls')
    url1 = 'http://staff.gemini.edu/~sleggett/2010_phot.xls'

    data_file2 = os.path.join(input_path, 'datafile8.txt')
    url2 = 'http://staff.gemini.edu/~sleggett/datafile8.txt'

    if not os.path.isfile(data_file1):
        sys.stdout.write('Downloading Leggett L and T Dwarf Data (88 kB)...')
        sys.stdout.flush()

        urlretrieve(url1, data_file1)

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    if not os.path.isfile(data_file2):
        sys.stdout.write('Downloading Leggett T6+ and Y Dwarf Data (44 kB)...')
        sys.stdout.flush()

        urlretrieve(url2, data_file2)

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    sys.stdout.write('Adding Leggett L and T Dwarf Data...')
    sys.stdout.flush()

    group = 'photometry/leggett'

    database.create_group(group)

    dataframe = pd.pandas.read_excel(data_file1)
    dataframe.columns = dataframe.columns.str.replace('\'', '')

    modulus = np.asarray(dataframe['M-m']) # M-m [mag]
    distance = 10.**(-modulus/5.+1.) # [pc]

    name = np.asarray(dataframe['Name'])

    sptype = np.asarray(dataframe['Type'])
    sptype = util.update_sptype(sptype)

    mag_y = np.asarray(dataframe['Y'])
    mag_j = np.asarray(dataframe['J'])
    mag_h = np.asarray(dataframe['H'])
    mag_k = np.asarray(dataframe['K'])
    mag_lp = np.asarray(dataframe['L'])
    mag_mp = np.asarray(dataframe['M'])
    mag_ch1 = np.asarray(dataframe['Ch1'])
    mag_ch2 = np.asarray(dataframe['Ch2'])
    mag_ch3 = np.asarray(dataframe['Ch3'])
    mag_ch4 = np.asarray(dataframe['Ch4'])
    mag_w1 = np.repeat(np.nan, np.size(name))
    mag_w2 = np.repeat(np.nan, np.size(name))
    mag_w3 = np.repeat(np.nan, np.size(name))
    mag_w4 = np.repeat(np.nan, np.size(name))

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    sys.stdout.write('Adding Leggett T6+ and Y Dwarf Data...')
    sys.stdout.flush()

    file_io = open(data_file2, 'r')
    lines = file_io.readlines()[69:]

    for item in lines:
        name = np.append(name, item[0:16])

        spt_tmp = item[62:66]
        if spt_tmp[0] == '2':
            spt_tmp = 'T'+spt_tmp[1]
        elif spt_tmp[0] == '3':
            spt_tmp = 'Y'+spt_tmp[1]

        sptype = np.append(sptype, spt_tmp)

        modulus = float(item[67:73]) # M-m [mag]
        if modulus == 999.:
            modulus = np.nan

        distance = np.append(distance, 10.**(-modulus/5.+1.)) # [pc]

        mag = np.zeros(14)

        mag[0] = float(item[95:101]) # MKO Y
        mag[1] = float(item[102:107]) # MKO J
        mag[2] = float(item[108:114]) # MKO H
        mag[3] = float(item[115:121]) # MKO K
        mag[4] = float(item[122:128]) # MKO L'
        mag[5] = float(item[129:135]) # MKO M'
        mag[6] = float(item[136:142]) # Spitzer/IRAC 3.6 micron
        mag[7] = float(item[143:149]) # Spitzer/IRAC 4.5 micron
        mag[8] = float(item[150:156]) # Spitzer/IRAC 5.8 micron
        mag[9] = float(item[157:163]) # Spitzer/IRAC 8.0 micron
        mag[10] = float(item[164:170]) # WISE W1
        mag[11] = float(item[171:176]) # WISE W2
        mag[12] = float(item[177:183]) # WISE W3
        mag[13] = float(item[184:190]) # WISE W4

        for j, mag_item in enumerate(mag):
            if mag_item == 999.:
                mag[j] = np.nan

        mag_y = np.append(mag_y, mag[0])
        mag_j = np.append(mag_j, mag[1])
        mag_h = np.append(mag_h, mag[2])
        mag_k = np.append(mag_k, mag[3])
        mag_lp = np.append(mag_lp, mag[4])
        mag_mp = np.append(mag_mp, mag[5])
        mag_ch1 = np.append(mag_ch1, mag[6])
        mag_ch2 = np.append(mag_ch2, mag[7])
        mag_ch3 = np.append(mag_ch3, mag[8])
        mag_ch4 = np.append(mag_ch4, mag[9])
        mag_w1 = np.append(mag_w1, mag[10])
        mag_w2 = np.append(mag_w2, mag[11])
        mag_w3 = np.append(mag_w3, mag[12])
        mag_w4 = np.append(mag_w4, mag[13])

    file_io.close()

    dtype = h5py.special_dtype(vlen=bytes)

    dset = database.create_dataset(group+'/name', (np.size(name), ), dtype=dtype)
    dset[...] = name

    dset = database.create_dataset(group+'/sptype', (np.size(sptype), ), dtype=dtype)
    dset[...] = sptype

    flag = np.repeat('null', np.size(name))

    dset = database.create_dataset(group+'/flag', (np.size(flag), ), dtype=dtype)
    dset[...] = flag

    database.create_dataset(group+'/distance', data=distance, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Y', data=mag_y, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.J', data=mag_j, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.H', data=mag_h, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.K', data=mag_k, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Lp', data=mag_lp, dtype='f')
    database.create_dataset(group+'/MKO/NSFCam.Mp', data=mag_mp, dtype='f')
    database.create_dataset(group+'/Spitzer/IRAC.I1', data=mag_ch1, dtype='f')
    database.create_dataset(group+'/Spitzer/IRAC.I2', data=mag_ch2, dtype='f')
    database.create_dataset(group+'/Spitzer/IRAC.I3', data=mag_ch3, dtype='f')
    database.create_dataset(group+'/Spitzer/IRAC.I4', data=mag_ch4, dtype='f')
    database.create_dataset(group+'/WISE/WISE.W1', data=mag_w1, dtype='f')
    database.create_dataset(group+'/WISE/WISE.W2', data=mag_w2, dtype='f')
    database.create_dataset(group+'/WISE/WISE.W3', data=mag_w3, dtype='f')
    database.create_dataset(group+'/WISE/WISE.W4', data=mag_w4, dtype='f')

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    database.close()



'''
Text
'''

import os
import sys

from urllib.request import urlretrieve

import numpy as np

from astropy.io import fits


def add_vega(input_path, database):
    '''
    Function for adding a Vega spectrum to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    data_file = os.path.join(input_path, 'alpha_lyr_stis_008.fits')
    url = 'http://ssb.stsci.edu/cdbs/calspec/alpha_lyr_stis_008.fits'

    if not os.path.isfile(data_file):
        sys.stdout.write('Downloading Vega spectrum (270 kB)... ')
        sys.stdout.flush()

        urlretrieve(url, data_file)

        sys.stdout.write('[DONE]\n')
        sys.stdout.flush()

    if 'spectra/calibration' not in database:
        database.create_group('spectra/calibration')

    if 'spectra/calibration/vega' in database:
        del database['spectra/calibration/vega']

    hdu = fits.open(data_file)
    data = hdu[1].data
    wavelength = data['WAVELENGTH'] # [Angstrom]
    flux = data['FLUX'] # [erg s-1 cm-2 A-1]
    error_stat = data['STATERROR'] # [erg s-1 cm-2 A-1]
    error_sys = data['SYSERROR'] # [erg s-1 cm-2 A-1]
    hdu.close()

    wavelength *= 1e-4 # [Angstrom] -> [micron]
    flux *= 1.e-3*1e4 # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]
    error_stat *= 1.e-3*1e4 # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]
    error_sys *= 1.e-3*1e4 # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]

    sys.stdout.write('Adding Vega spectrum... ')
    sys.stdout.flush()

    database.create_dataset('spectra/calibration/vega',
                            data=np.vstack((wavelength, flux, error_stat)),
                            dtype='f')

    sys.stdout.write('[DONE]\n')
    sys.stdout.flush()



'''
Read module.
'''

import os
import math
import configparser

import h5py
import numpy as np

from species.core import box
from species.core import constants as con
from species.data import database
from species.read import read_filter


def get_planck(temperature,
               radius,
               distance,
               wavelength,
               specres):
    '''
    :param temperature: Temperature (K)
    :type temperature: float
    :param radius: Radius (Rjup).
    :type radius: float
    :param distance: Distance (pc).
    :type distance: float
    :param wavelength: Wavelength range (micron).
    :type wavelength: tuple(float, float)
    :param specres: Spectral resolution
    :type specres: float

    :return: Box with the Planck spectrum.
    :rtype: species.core.box.SpectrumBox
    '''

    wl_points = [wavelength[0]]

    while wl_points[-1] <= wavelength[1]:
        wl_points.append(wl_points[-1] + wl_points[-1]/specres)

    wl_points = np.asarray(wl_points[:-1]) # [micron]
    wl_points *= 1e-6 # [m]

    planck1 = 2.*con.PLANCK*con.LIGHT**2/wl_points**5
    planck2 = np.exp(con.PLANCK*con.LIGHT/(wl_points*con.BOLTZMANN*temperature)) - 1.

    flux = 4.*math.pi * (radius*con.R_JUP/(distance*con.PARSEC))**2 * planck1/planck2 # [W m-2 m-1]
    flux *= 1e-6 # [W m-2 micron-1]

    wl_points *= 1e6 # [micron]

    return box.create_box(boxtype='spectrum',
                          spectrum='planck',
                          wavelength=wl_points,
                          flux=flux,
                          name=None)


class ReadSpectrum:
    '''
    Text
    '''

    def __init__(self,
                 spectrum,
                 filter_name):
        '''
        :param spectrum: Spectral library.
        :type spectrum: str
        :param filter_name: Filter name. Full spectrum is read if set to None.
        :type filter_name: str

        :return: None
        '''

        self.spectrum = spectrum
        self.filter_name = filter_name

        if filter_name is None:
            self.wl_range = None

        else:
            transmission = read_filter.ReadFilter(filter_name)
            self.wl_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def get_spectrum(self,
                     ignore_nan=True,
                     sptype=None):
        '''
        :param ignore_nan: Ignore wavelength points for which the flux is NaN.
        :type ignore_nan: bool
        :param sptype: Spectral types to select. All spectra are retrieved if set to None.
        :type sptype: str

        :return: Box with the spectra.
        :rtype: species.core.box.SpectrumBox
        '''

        h5_file = h5py.File(self.database, 'r')

        try:
            h5_file['spectra/'+self.spectrum]

        except KeyError:
            h5_file.close()
            species_db = database.Database()
            species_db.add_spectrum(self.spectrum)
            h5_file = h5py.File(self.database, 'r')

        list_wavelength = []
        list_flux = []
        list_name = []
        list_simbad = []
        list_sptype = []
        list_distance = []

        for item in h5_file['spectra/'+self.spectrum]:
            data = h5_file['spectra/'+self.spectrum+'/'+item]

            wavelength = data[0, :] # [micron]
            flux = data[1, :] # [W m-2 micron-1]

            if data.shape[0] == 3:
                error = data[2, :] # [W m-2 micron-1]

            if ignore_nan:
                indices = np.isnan(flux)
                indices = np.logical_not(indices)
                indices = np.where(indices)[0]

                wavelength = wavelength[indices]
                flux = flux[indices]

                if data.shape[0] == 3:
                    error = error[indices]

            if self.wl_range is None:
                wl_index = np.arange(0, len(wavelength), 1)

            else:
                wl_index = (flux > 0.) & (wavelength > self.wl_range[0]) & \
                           (wavelength < self.wl_range[1])

            count = np.count_nonzero(wl_index)

            if count > 0:
                index = np.where(wl_index)[0]

                if index[0] > 0:
                    wl_index[index[0] - 1] = True

                if index[-1] < len(wl_index)-1:
                    wl_index[index[-1] + 1] = True

                list_wavelength.append(wavelength[wl_index])
                list_flux.append(flux[wl_index])

                attrs = data.attrs
                if 'name' in attrs:
                    list_name.append(data.attrs['name'])

                if 'simbad' in attrs:
                    list_simbad.append(data.attrs['simbad'])

                if 'sptype' in attrs:
                    list_sptype.append(data.attrs['sptype'])

                if 'distance' in attrs:
                    list_distance.append(data.attrs['distance'])

        specbox = box.SpectrumBox()

        specbox.spectrum = self.spectrum
        specbox.wavelength = np.asarray(list_wavelength)
        specbox.flux = np.asarray(list_flux)

        if list_name:
            specbox.name = np.asarray(list_name)

        if list_simbad:
            specbox.simbad = np.asarray(list_simbad)

        if list_sptype:
            specbox.sptype = np.asarray(list_sptype)

        if list_distance:
            specbox.distance = np.asarray(list_distance)

        if sptype is not None:
            indices = np.where(np.chararray.startswith(specbox.sptype, sptype))[0]

            specbox.wavelength = specbox.wavelength[indices]
            specbox.flux = specbox.flux[indices]
            specbox.name = specbox.name[indices]
            specbox.simbad = specbox.simbad[indices]
            specbox.sptype = specbox.sptype[indices]
            specbox.distance = specbox.distance[indices]

        return specbox



'''
Text
'''

import os
import math
import configparser

import h5py
import numpy as np

from scipy.integrate import simps
from scipy.interpolate import RegularGridInterpolator
from PyAstronomy.pyasl import instrBroadGaussFast

from species.analysis import photometry
from species.core import box, constants
from species.data import database
from species.read import read_filter


def get_mass(model_par):
    '''
    :param model_par: Model parameter values. Should contain the surface gravity and radius.
    :type model_par: dict

    :return: Mass (Mjup).
    :rtype: float
    '''

    logg = 1e-2 * 10.**model_par['logg'] # [m s-1]

    radius = model_par['radius'] # [Rjup]
    radius *= constants.R_JUP # [m]

    mass = logg*radius**2/constants.GRAVITY # [kg]
    mass /= constants.M_JUP # [Mjup]

    return mass


def add_luminosity(modelbox,
                   specres=1000):
    '''
    Function to add the luminosity of a model spectrum to the parameter dictionary of the box. The
    luminosity is by default calculated at a spectral resolution of 1000.

    :param modelbox: Box with the model spectrum. Should also contain the dictionary with the model
                     parameters, the radius in particular.
    :type modelbox: species.core.box.ModelBox
    :param specres: Spectral resolution of the interpolated spectrum.
    :type specres: float

    :return: The input box with the luminosity added in the parameter dictionary.
    :rtype: species.core.box.ModelBox
    '''

    readmodel = ReadModel(model=modelbox.model, wavelength=None, teff=None)
    fullspec = readmodel.get_model(model_par=modelbox.parameters, sampling=('specres', specres))

    flux = simps(fullspec.flux, fullspec.wavelength)

    if 'distance' in modelbox.parameters:
        luminosity = 4.*math.pi*(fullspec.parameters['distance']*constants.PARSEC)**2*flux # [W]
    else:
        luminosity = 4.*math.pi*(fullspec.parameters['radius']*constants.R_JUP)**2*flux # [W]

    modelbox.parameters['luminosity'] = luminosity/constants.L_SUN # [Lsun]

    return modelbox


class ReadModel:
    '''
    Text
    '''

    def __init__(self,
                 model,
                 wavelength,
                 teff=None):
        '''
        :param model: Model name.
        :type model: str
        :param wavelength: Wavelength range (micron) or filter name. Full spectrum if set to None.
        :type wavelength: tuple(float, float) or str
        :param teff: Effective temperature (K) range. Restricting the temperature range will speed
                     up the computation.
        :type teff: tuple(float, float)

        :return: None
        '''


        self.model = model
        self.teff = teff
        self.spectrum_interp = None

        if isinstance(wavelength, str):
            self.filter_name = wavelength
            transmission = read_filter.ReadFilter(wavelength)
            self.wavelength = transmission.wavelength_range()

        else:
            self.filter_name = None
            self.wavelength = wavelength

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def open_database(self):
        '''
        :return: Database.
        :rtype: h5py._hl.files.File
        '''

        h5_file = h5py.File(self.database, 'r')

        try:
            h5_file['models/'+self.model]

        except KeyError:
            h5_file.close()
            species_db = database.Database()
            species_db.add_model(self.model, self.wavelength, self.teff)
            h5_file = h5py.File(self.database, 'r')

        return h5_file

    def interpolate(self):
        '''
        :return: None
        '''

        h5_file = self.open_database()

        wavelength = np.asarray(h5_file['models/'+self.model+'/wavelength'])
        flux = np.asarray(h5_file['models/'+self.model+'/flux'])
        teff = np.asarray(h5_file['models/'+self.model+'/teff'])
        logg = np.asarray(h5_file['models/'+self.model+'/logg'])

        wl_index = (wavelength > self.wavelength[0]) & (wavelength < self.wavelength[1])

        index = np.where(wl_index)[0]

        wl_index[index[0] - 1] = True
        wl_index[index[-1] + 1] = True

        if self.model in ('drift-phoenix', 'bt-nextgen'):
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])
            flux = flux[:, :, :, wl_index]
            points = (teff, logg, feh, wavelength[wl_index])

        # elif self.model == 'petitcode_warm_clear':
        #     feh = np.asarray(h5_file['models/petitcode_warm_clear/feh'])
        #     flux = flux[:, :, :, wl_index]
        #     points = np.asarray((teff, logg, feh, wavelength[wl_index]))
        #
        # elif self.model == 'petitcode_warm_cloudy':
        #     feh = np.asarray(h5_file['models/petitcode_warm_cloudy/feh'])
        #     fsed = np.asarray(h5_file['models/petitcode_warm_cloudy/fsed'])
        #     flux = flux[:, :, :, :, wl_index]
        #     points = np.asarray((teff, logg, feh, fsed, wavelength[wl_index]))
        #
        # elif self.model == 'petitcode_hot_clear':
        #     feh = np.asarray(h5_file['models/petitcode_hot_clear/feh'])
        #     co_ratio = np.asarray(h5_file['models/petitcode_hot_clear/co'])
        #     flux = flux[:, :, :, :, wl_index]
        #     points = np.asarray((teff, logg, feh, co_ratio, wavelength[wl_index]))
        #
        # elif self.model == 'petitcode_hot_cloudy':
        #     feh = np.asarray(h5_file['models/petitcode_hot_cloudy/feh'])
        #     co_ratio = np.asarray(h5_file['models/petitcode_hot_cloudy/co'])
        #     fsed = np.asarray(h5_file['models/petitcode_hot_cloudy/fsed'])
        #     flux = flux[:, :, :, :, :, wl_index]
        #     points = np.asarray((teff, logg, feh, co_ratio, fsed, wavelength[wl_index]))

        h5_file.close()

        self.spectrum_interp = RegularGridInterpolator(points=points,
                                                       values=flux,
                                                       method='linear',
                                                       bounds_error=False,
                                                       fill_value=np.nan)

    def get_data(self,
                 model_par):
        '''
        :param model_par: Model parameter values. Only discrete values from the original grid
                          are possible. Else, the nearest grid values are selected.
        :type model_par: dict

        :return: Spectrum (micron, W m-2 micron-1).
        :rtype: species.core.box.ModelBox
        '''

        h5_file = self.open_database()

        wavelength = np.asarray(h5_file['models/'+self.model+'/wavelength'])
        flux = np.asarray(h5_file['models/'+self.model+'/flux'])
        teff = np.asarray(h5_file['models/'+self.model+'/teff'])
        logg = np.asarray(h5_file['models/'+self.model+'/logg'])

        if self.wavelength is None:
            wl_index = np.ones(wavelength.shape[0], dtype=bool)

        else:
            wl_index = (wavelength > self.wavelength[0]) & (wavelength < self.wavelength[1])
            index = np.where(wl_index)[0]

            wl_index[index[0] - 1] = True
            wl_index[index[-1] + 1] = True

        if self.model in ('drift-phoenix', 'bt-nextgen'):
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])

            teff_index = np.argwhere(teff == model_par['teff'])[0]

            if not teff_index:
                raise ValueError('Temperature value not found.')
            else:
                teff_index = teff_index[0]

            logg_index = np.argwhere(logg == model_par['logg'])[0]

            if not logg_index:
                raise ValueError('Surface gravity value not found.')
            else:
                logg_index = logg_index[0]

            feh_index = np.argwhere(feh == model_par['feh'])[0]

            if not feh_index:
                raise ValueError('Metallicity value not found.')
            else:
                feh_index = feh_index[0]

            wavelength = wavelength[wl_index]
            flux = flux[teff_index, logg_index, feh_index, wl_index]

        if 'radius' in model_par and 'distance' in model_par:
            scaling = (model_par['radius']*constants.R_JUP)**2 / \
                      (model_par['distance']*constants.PARSEC)**2
            flux *= scaling

        return box.create_box(boxtype='model',
                              model=self.model,
                              wavelength=wavelength,
                              flux=flux,
                              parameters=model_par)

    def get_model(self,
                  model_par,
                  sampling):
        '''
        :param model_par: Model parameter values.
        :type model_par: dict
        :param sampling: Type of wavelength sampling.
        :type sampling: tuple

        :return: Spectrum (micron, W m-2 micron-1).
        :rtype: species.core.box.ModelBox
        '''

        if not self.wavelength:
            wl_points = self.get_wavelength()
            self.wavelength = (wl_points[0], wl_points[-1])

        if self.spectrum_interp is None:
            self.interpolate()

        wavelength = [self.wavelength[0]]

        if sampling[0] == 'specres':
            while wavelength[-1] <= self.wavelength[1]:
                wavelength.append(wavelength[-1] + wavelength[-1]/sampling[1])
            wavelength = np.asarray(wavelength[:-1])

        elif sampling[0] == 'gaussian':
            wavelength = np.linspace(self.wavelength[0], self.wavelength[1], sampling[1][0])

        flux = np.zeros(wavelength.shape)

        for i, item in enumerate(wavelength):

            if self.model in ('drift-phoenix', 'bt-nextgen'):
                parameters = [model_par['teff'],
                              model_par['logg'],
                              model_par['feh'],
                              item]

            # elif self.model == 'petitcode_warm_clear':
            #     parameters = [model_par['teff'],
            #                   model_par['logg'],
            #                   model_par['feh'],
            #                   item]
            #
            # elif self.model == 'petitcode_warm_cloudy':
            #     parameters = [model_par['teff'],
            #                   model_par['logg'],
            #                   model_par['feh'],
            #                   model_par['fsed'],
            #                   item]
            #
            # elif self.model == 'petitcode_hot_clear':
            #     parameters = [model_par['teff'],
            #                   model_par['logg'],
            #                   model_par['feh'],
            #                   model_par['co'],
            #                   item]
            #
            # elif self.model == 'petitcode_hot_cloudy':
            #     parameters = [model_par['teff'],
            #                   model_par['logg'],
            #                   model_par['feh'],
            #                   model_par['co'],
            #                   model_par['fsed'],
            #                   item]

            flux[i] = self.spectrum_interp(np.asarray(parameters))

        if 'radius' in model_par:
            model_par['mass'] = get_mass(model_par)

            if 'distance' in model_par:
                scaling = (model_par['radius']*constants.R_JUP)**2 / \
                          (model_par['distance']*constants.PARSEC)**2
                flux *= scaling

        if sampling[0] == 'gaussian':
            index = np.where(np.isnan(flux))[0]

            wavelength = np.delete(wavelength, index)
            flux = np.delete(flux, index)

            flux = instrBroadGaussFast(wavelength, flux, sampling[1][1])

        return box.create_box(boxtype='model',
                              model=self.model,
                              wavelength=wavelength,
                              flux=flux,
                              parameters=model_par)

    def get_photometry(self,
                       model_par,
                       sampling,
                       synphot=None):
        '''
        :param model_par: Model parameter values.
        :type model_par: dict
        :param sampling: Spectral sampling. The original grid is used (nearest model parameter
                         values) if set to none.
        :type sampling: float
        :param synphot: Synthetic photometry object.
        :type synphot: species.analysis.photometry.SyntheticPhotometry

        :return: Average flux density (W m-2 micron-1).
        :rtype: float
        '''

        if sampling is None:
            spectrum = self.get_data(model_par)

        else:
            if self.spectrum_interp is None:
                self.interpolate()

            spectrum = self.get_model(model_par, sampling)

        if not synphot:
            synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_photometry(spectrum.wavelength, spectrum.flux)

    # def get_magnitude(self, model_par, sampling):
    #     '''
    #     :param model_par: Model parameter values.
    #     :type model_par: dict
    #     :param sampling: Spectral sampling. The original grid is used (nearest model parameter
    #                     values) if set to none.
    #     :type sampling: float
    #
    #     :return: Apparent magnitude (mag), absolute magnitude (mag).
    #     :rtype: float, float
    #     '''
    #
    #     if sampling is None:
    #         spectrum = self.get_data(model_par)
    #
    #     else:
    #         if self.spectrum_interp is None:
    #             self.interpolate()
    #
    #         spectrum = self.get_model(model_par, sampling)
    #
    #     transmission = read_filter.ReadFilter(self.filter_name)
    #     filter_interp = transmission.interpolate()
    #
    #     synphot = photometry.SyntheticPhotometry(filter_interp)
    #     mag = synphot.spectrum_to_magnitude(spectrum.wavelength,
    #                                         spectrum.flux,
    #                                         model_par['distance'])
    #
    #     return mag[0], mag[1]

    def get_bounds(self):
        '''
        :return: Parameter boundaries of the model grid.
        :rtype: dict
        '''

        h5_file = self.open_database()

        teff = h5_file['models/'+self.model+'/teff']
        logg = h5_file['models/'+self.model+'/logg']

        if self.model in ('drift-phoenix', 'bt-nextgen'):
            feh = h5_file['models/'+self.model+'/feh']
            bounds = {'teff':(teff[0], teff[-1]),
                      'logg':(logg[0], logg[-1]),
                      'feh':(feh[0], feh[-1])}

        h5_file.close()

        return bounds

    def get_wavelength(self):
        '''
        :return: Wavelength points (micron).
        :rtype: numpy.ndarray
        '''

        h5_file = self.open_database()
        wavelength = np.asarray(h5_file['models/'+self.model+'/wavelength'])
        h5_file.close()

        return wavelength

    def get_points(self):
        '''
        :return: Parameter points of the model grid.
        :rtype: dict
        '''

        points = {}

        h5_file = self.open_database()

        teff = h5_file['models/'+self.model+'/teff']
        logg = h5_file['models/'+self.model+'/logg']

        points['teff'] = np.asarray(teff)
        points['logg'] = np.asarray(logg)

        if self.model == 'drift-phoenix':
            feh = h5_file['models/'+self.model+'/feh']
            points['feh'] = np.asarray(feh)

        h5_file.close()

        return points

    def get_parameters(self):
        '''
        :return: Model parameters.
        :rtype: list(str, )
        '''

        h5_file = self.open_database()

        dset = h5_file['models/'+self.model]
        nparam = dset.attrs['nparam']

        param = []
        for i in range(nparam):
            param.append(dset.attrs['parameter'+str(i)])

        return param



'''
Read module.
'''

import os
import configparser

import h5py
import numpy as np

from species.analysis import photometry
from species.core import box
from species.data import database


class ReadColorMagnitude:
    '''
    Text
    '''

    def __init__(self,
                 library,
                 filters_color,
                 filter_mag):
        '''
        :param library: Photometric libraries.
        :type library: tuple(str, )
        :param filters_color:
        :type filters_color: tuple(str, str)
        :param filter_mag:
        :type filter_mag: str

        :return: None
        '''

        self.library = library
        self.filters_color = filters_color
        self.filter_mag = filter_mag

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        if isinstance(self.library, str):
            self.library = (self.library, )

    def get_color_magnitude(self,
                            object_type):
        '''
        :param object_type:
        :type object_type: str

        :return: Box with the colors and magnitudes.
        :rtype: species.core.box.ColorMagBox
        '''

        h5_file = h5py.File(self.database, 'r')

        for i, item in enumerate(self.library):
            try:
                h5_file['photometry/'+item]

            except KeyError:
                h5_file.close()
                species_db = database.Database()
                species_db.add_photometry(item)
                h5_file = h5py.File(self.database, 'r')

            sptype_tmp = np.asarray(h5_file['photometry/'+item+'/sptype'])
            distance_tmp = np.asarray(h5_file['photometry/'+item+'/distance']) # [pc]
            flag_tmp = np.asarray(h5_file['photometry/'+item+'/flag'])

            if object_type == 'field':
                indices_tmp = np.where(flag_tmp == b'null')[0]
            else:
                indices_tmp = np.arange(0, np.size(sptype_tmp), 1)

            if i == 0:
                sptype = sptype_tmp
                distance = distance_tmp
                flag = flag_tmp
                indices = indices_tmp

                mag1 = np.asarray(h5_file['photometry/'+item+'/'+self.filters_color[0]])
                mag2 = np.asarray(h5_file['photometry/'+item+'/'+self.filters_color[1]])

            else:
                distance_tmp = np.asarray(h5_file['photometry/'+item+'/distance']) # [pc]
                distance = np.concatenate((distance, distance_tmp), axis=0)

                sptype_tmp = np.asarray(h5_file['photometry/'+item+'/sptype'])
                sptype = np.concatenate((sptype, sptype_tmp), axis=0)

                flag = np.concatenate((flag, flag_tmp), axis=0)
                indices = np.concatenate((indices, indices.shape+indices_tmp), axis=0)

                mag1_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters_color[0]])
                mag2_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters_color[1]])

                mag1 = np.concatenate((mag1, mag1_tmp), axis=0)
                mag2 = np.concatenate((mag2, mag2_tmp), axis=0)

        color = mag1 - mag2

        if self.filter_mag == self.filters_color[0]:
            mag = photometry.apparent_to_absolute(mag1, distance)

        elif self.filter_mag == self.filters_color[1]:
            mag = photometry.apparent_to_absolute(mag2, distance)

        h5_file.close()

        return box.create_box(boxtype='colormag',
                              library=self.library,
                              object_type=object_type,
                              filters_color=self.filters_color,
                              filter_mag=self.filter_mag,
                              color=color[indices],
                              magnitude=mag[indices],
                              sptype=sptype[indices])


class ReadColorColor:
    '''
    Text
    '''

    def __init__(self,
                 library,
                 filters):
        '''
        :param library: Photometric libraries.
        :type library: tuple(str, )
        :param filters: Filter IDs for the two colors.
        :type filters: tuple(tuple(str, str), tuple(str, str))

        :return: None
        '''

        self.library = library
        self.filters = filters

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        if isinstance(self.library, str):
            self.library = (self.library, )

    def get_color_color(self,
                        object_type):
        '''
        :param object_type: Object type (currently only 'field' possible). All objects are
                            selected if set to None.
        :type object_type: str

        :return: Box with the colors.
        :rtype: species.core.box.ColorColorBox
        '''

        h5_file = h5py.File(self.database, 'r')

        for i, item in enumerate(self.library):
            try:
                h5_file['photometry/'+item]

            except KeyError:
                h5_file.close()
                species_db = database.Database()
                species_db.add_photometry(item)
                h5_file = h5py.File(self.database, 'r')

            sptype_tmp = np.asarray(h5_file['photometry/'+item+'/sptype'])
            distance_tmp = np.asarray(h5_file['photometry/'+item+'/distance']) # [pc]
            flag_tmp = np.asarray(h5_file['photometry/'+item+'/flag'])

            if object_type == 'field':
                indices_tmp = np.where(flag_tmp == b'null')[0]
            else:
                indices_tmp = np.arange(0, np.size(sptype_tmp), 1)

            if i == 0:
                sptype = sptype_tmp
                distance = distance_tmp
                flag = flag_tmp
                indices = indices_tmp

                mag1 = np.asarray(h5_file['photometry/'+item+'/'+self.filters[0][0]])
                mag2 = np.asarray(h5_file['photometry/'+item+'/'+self.filters[0][1]])
                mag3 = np.asarray(h5_file['photometry/'+item+'/'+self.filters[1][0]])
                mag4 = np.asarray(h5_file['photometry/'+item+'/'+self.filters[1][1]])

            else:
                distance_tmp = np.asarray(h5_file['photometry/'+item+'/distance']) # [pc]
                distance = np.concatenate((distance, distance_tmp), axis=0)

                sptype_tmp = np.asarray(h5_file['photometry/'+item+'/sptype'])
                sptype = np.concatenate((sptype, sptype_tmp), axis=0)

                flag = np.concatenate((flag, flag_tmp), axis=0)
                indices = np.concatenate((indices, indices.shape+indices_tmp), axis=0)

                mag1_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters[0][0]])
                mag2_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters[0][1]])
                mag3_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters[1][0]])
                mag4_tmp = np.asarray(h5_file['photometry/'+item+'/'+self.filters[1][1]])

                mag1 = np.concatenate((mag1, mag1_tmp), axis=0)
                mag2 = np.concatenate((mag2, mag2_tmp), axis=0)
                mag3 = np.concatenate((mag3, mag3_tmp), axis=0)
                mag4 = np.concatenate((mag4, mag4_tmp), axis=0)

        color1 = mag1 - mag2
        color2 = mag3 - mag4

        h5_file.close()

        return box.create_box(boxtype='colorcolor',
                              library=self.library,
                              object_type=object_type,
                              filters=self.filters,
                              color1=color1[indices],
                              color2=color2[indices],
                              sptype=sptype[indices])



'''
Module with functions for making plots.
'''

import os
import sys
import math
import itertools

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from matplotlib.colorbar import Colorbar

from species.plot import util
from species.read import read_object


mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
mpl.rcParams['font.family'] = 'serif'

plt.rc('axes', edgecolor='black', linewidth=2)

marker = itertools.cycle(('o', 's', 'p', '<', '>', 'p', 'v', '^', '*', 'd', 'x', '+', '1', '2', '3', '4'))

def plot_color_magnitude(colorbox,
                         objects,
                         label_x,
                         label_y,
                         output,
                         xlim=None,
                         ylim=None,
                         offset=None,
                         legend='top left'):
    '''
    :param colorbox: Box with the colors and magnitudes.
    :type colorbox: species.core.box.ColorMagBox
    :param objects: Tuple with individual objects. The objects require a tuple with their database
                    tag, the two filter IDs for the color, and the filter ID for the absolute
                    magnitude.
    :type objects: tuple(tuple(str, str, str, str), )
    :param label_x: Label for the x-axis.
    :type label_x: str
    :param label_y: Label for the y-axis.
    :type label_y: str
    :param output: Output filename.
    :type output: str
    :param xlim: Limits for the x-axis.
    :type xlim: tuple(float, float)
    :param ylim: Limits for the y-axis.
    :type ylim: tuple(float, float)
    :param legend: Legend position.
    :type legend: str

    :return: None
    '''

    sys.stdout.write('Plotting color-magnitude diagram: '+output+'... ')
    sys.stdout.flush()

    plt.figure(1, figsize=(4, 4.8))
    gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.5])
    gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[2, 0])
    ax2 = plt.subplot(gridsp[0, 0])

    ax1.grid(True, linestyle=':', linewidth=0.7, color='silver', dashes=(1, 4), zorder=0)

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

    cmap = plt.cm.viridis
    bounds = np.arange(0, 8, 1)
    norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

    sptype = colorbox.sptype
    color = colorbox.color
    magnitude = colorbox.magnitude

    indices = np.where(sptype != 'None')[0]

    sptype = sptype[indices]
    color = color[indices]
    magnitude = magnitude[indices]

    spt_disc = util.sptype_discrete(sptype, color.shape)

    _, unique = np.unique(color, return_index=True)

    sptype = sptype[unique]
    color = color[unique]
    magnitude = magnitude[unique]
    spt_disc = spt_disc[unique]

    scat = ax1.scatter(color, magnitude, c=spt_disc, cmap=cmap, norm=norm,
                       zorder=1, s=25., alpha=0.6)

    cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                  ticklocation='top', format='%.2f')

    cb.ax.tick_params(width=0.8, length=5, labelsize=10, direction='in', color='white')
    cb.set_ticks(np.arange(0.5, 7., 1.))
    cb.set_ticklabels(['M0-M4', 'M5-M9', 'L0-L4', 'L5-L9', 'T0-T4', 'T6-T8', 'Y1-Y2'])

    if objects is not None:
        for item in objects:
            objdata = read_object.ReadObject(item[0])

            objcolor1 = objdata.get_photometry(item[1])
            objcolor2 = objdata.get_photometry(item[2])
            abs_mag = objdata.get_absmag(item[3])

            colorerr = math.sqrt(objcolor1[1]**2+objcolor2[1]**2)

            ax1.errorbar(objcolor1[0]-objcolor2[0], abs_mag[0], yerr=abs_mag[1], xerr=colorerr,
                         marker=next(marker), ms=6, color='black', label=objdata.object_name,
                         markerfacecolor='white', markeredgecolor='black', zorder=2)

    handles, labels = ax1.get_legend_handles_labels()

    if handles:
        handles = [h[0] for h in handles]
        ax1.legend(handles, labels, loc=legend, prop={'size':9}, frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.close()

    sys.stdout.write('[DONE]\n')
    sys.stdout.flush()


def plot_color_color(colorbox,
                     objects,
                     label_x,
                     label_y,
                     output,
                     xlim=None,
                     ylim=None,
                     offset=None,
                     legend='top left'):
    '''
    :param colorbox: Box with the colors and magnitudes.
    :type colorbox: species.core.box.ColorMagBox
    :param objects: Tuple with individual objects. The objects require a tuple with their database
                    tag, the two filter IDs for the color, and the filter ID for the absolute
                    magnitude.
    :type objects: tuple(tuple(str, str, str, str), )
    :param label_x: Label for the x-axis.
    :type label_x: str
    :param label_y: Label for the y-axis.
    :type label_y: str
    :param output: Output filename.
    :type output: str
    :param xlim: Limits for the x-axis.
    :type xlim: tuple(float, float)
    :param ylim: Limits for the y-axis.
    :type ylim: tuple(float, float)
    :param offset: Offset of the x- and y-axis label.
    :type offset: tuple(float, float)
    :param legend: Legend position.
    :type legend: str

    :return: None
    '''

    sys.stdout.write('Plotting color-color diagram: '+output+'... ')
    sys.stdout.flush()

    plt.figure(1, figsize=(4, 4.3))
    gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.])
    gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[2, 0])
    ax2 = plt.subplot(gridsp[0, 0])

    ax1.grid(True, linestyle=':', linewidth=0.7, color='silver', dashes=(1, 4), zorder=0)

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

    cmap = plt.cm.viridis
    bounds = np.arange(0, 8, 1)
    norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

    sptype = colorbox.sptype
    color1 = colorbox.color1
    color2 = colorbox.color2

    indices = np.where(sptype != 'None')[0]

    sptype = sptype[indices]
    color1 = color1[indices]
    color2 = color2[indices]

    spt_disc = util.sptype_discrete(sptype, color1.shape)

    _, unique = np.unique(color1, return_index=True)

    sptype = sptype[unique]
    color1 = color1[unique]
    color2 = color2[unique]
    spt_disc = spt_disc[unique]

    scat = ax1.scatter(color1, color2, c=spt_disc, cmap=cmap, norm=norm,
                       zorder=1, s=25., alpha=0.6)

    cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                  ticklocation='top', format='%.2f')

    cb.ax.tick_params(width=0.8, length=5, labelsize=10, direction='in', color='white')
    cb.set_ticks(np.arange(0.5, 7., 1.))
    cb.set_ticklabels(['M0-M4', 'M5-M9', 'L0-L4', 'L5-L9', 'T0-T4', 'T6-T8', 'Y1-Y2'])

    if objects is not None:
        for item in objects:
            objdata = read_object.ReadObject(item[0])

            mag1 = objdata.get_photometry(item[1][0])[0]
            mag2 = objdata.get_photometry(item[1][1])[0]
            mag3 = objdata.get_photometry(item[2][0])[0]
            mag4 = objdata.get_photometry(item[2][1])[0]

            err1 = objdata.get_photometry(item[1][0])[1]
            err2 = objdata.get_photometry(item[1][1])[1]
            err3 = objdata.get_photometry(item[2][0])[1]
            err4 = objdata.get_photometry(item[2][1])[1]

            color1 = mag1 - mag2
            color2 = mag3 - mag4

            error1 = math.sqrt(err1**2+err2**2)
            error2 = math.sqrt(err3**2+err4**2)

            ax1.errorbar(color1, color2, xerr=error1, yerr=error2,
                         marker=next(marker), ms=6, color='black', label=objdata.object_name,
                         markerfacecolor='white', markeredgecolor='black', zorder=2)

    handles, labels = ax1.get_legend_handles_labels()

    if handles:
        handles = [h[0] for h in handles]
        ax1.legend(handles, labels, loc=legend, prop={'size':9}, frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.close()

    sys.stdout.write('[DONE]\n')
    sys.stdout.flush()



import os
import shutil

import numpy as np

import species
from . import tools


class TestPhotometry:

    def setup_class(self):
        self.limit = 1e-10

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        shutil.rmtree("data/")

    def test_species_init(self):
        tools.create_config()
        species.SpeciesInit("./")

    def test_synthetic_photometry(self):
        species.SyntheticPhotometry("MKO/NSFCam.J")

    def test_magnitude_to_flux(self):
        synphot = species.SyntheticPhotometry("MKO/NSFCam.J")
        flux, error = synphot.magnitude_to_flux(20., 0.5)

        assert np.allclose(flux, 3.104539116258e-17, rtol=self.limit, atol=0.)
        assert np.allclose(error, 1.4807655772287564e-17, rtol=self.limit, atol=0.)

    def test_flux_to_magnitude(self):
        synphot = species.SyntheticPhotometry("MKO/NSFCam.J")
        app_mag, abs_mag = synphot.flux_to_magnitude(1e-10, 50.)

        assert np.allclose(app_mag, 3.72999284049042, rtol=self.limit, atol=0.)
        assert np.allclose(abs_mag, 0.23514281881032595, rtol=self.limit, atol=0.)



'''
Text.
'''

import numpy as np


def sptype_discrete(sptype,
                    shape):
    '''
    :param sptype:
    :type sptype:
    :param shape:
    :type shape:

    :return::
    :rtype: numpy.ndarray
    '''

    spt_disc = np.zeros(shape)

    for i, item in enumerate(sptype):
        sp = item[0:2]

        if sp in (np.string_('M0'), np.string_('M1'), np.string_('M2'), np.string_('M3'), np.string_('M4')):
            spt_disc[i] = 0

        elif sp in (np.string_('M5'), np.string_('M6'), np.string_('M7'), np.string_('M8'), np.string_('M9')):
            spt_disc[i] = 1

        elif sp in (np.string_('L0'), np.string_('L1'), np.string_('L2'), np.string_('L3'), np.string_('L4')):
            spt_disc[i] = 2

        elif sp in (np.string_('L5'), np.string_('L6'), np.string_('L7'), np.string_('L8'), np.string_('L9')):
            spt_disc[i] = 3

        elif sp in (np.string_('T0'), np.string_('T1'), np.string_('T2'), np.string_('T3'), np.string_('T4')):
            spt_disc[i] = 4

        elif sp in (np.string_('T5'), np.string_('T6'), np.string_('T7'), np.string_('T8'), np.string_('T9')):
            spt_disc[i] = 5

        elif np.string_('Y') in item:
            spt_disc[i] = 6

        else:
            spt_disc[i] = np.nan
            continue

    return spt_disc


def update_labels(param):
    '''
    :param param:
    :type param: list

    :return:
    :rtype: list
    '''

    if 'teff' in param:
        index = param.index('teff')
        param[index] = r'$T_\mathregular{eff}$ [K]'

    if 'logg' in param:
        index = param.index('logg')
        param[index] = r'$\log\,g$ [dex]'

    if 'feh' in param:
        index = param.index('feh')
        param[index] = r'Fe/H [dex]'

    if 'radius' in param:
        index = param.index('radius')
        param[index] = r'$R$ [$\mathregular{R_{Jup}}$]'

    if 'scaling' in param:
        index = param.index('scaling')
        param[index] = 'Scaling'

    if 'offset' in param:
        index = param.index('offset')
        param[index] = 'Offset'

    return param


def model_name(key):
    '''
    :param key:
    :type key: str

    :return:
    :rtype: str
    '''

    if key == 'drift-phoenix':
        name = 'DRIFT-PHOENIX'

    elif key == 'bt-nextgen':
        name = 'BT-NextGen'

    return name


def quantity_unit(param):
    '''
    :param param:
    :type param: list

    :return:
    :rtype: list, list
    '''

    quantity = []
    unit = []

    if 'teff' in param:
        quantity.append(r'$T_\mathregular{eff}$')
        unit.append('K')

    if 'logg' in param:
        quantity.append(r'$\log\,g$')
        unit.append('dex')

    if 'feh' in param:
        quantity.append(r'[Fe/H]')
        unit.append('dex')

    if 'radius' in param:
        quantity.append(r'$R$')
        unit.append(r'$R_\mathregular{Jup}}$')

    if 'distance' in param:
        quantity.append(r'$d$')
        unit.append('pc')

    if 'mass' in param:
        quantity.append(r'$M$')
        unit.append(r'$M_\mathregular{Jup}$')

    if 'luminosity' in param:
        quantity.append(r'$L$')
        unit.append(r'$L_\odot$')

    return quantity, unit



'''
Text
'''

import os
import sys
import tarfile

from urllib.request import urlretrieve

import numpy as np

from astropy.io import fits

from species.data import queries, util


def add_irtf(input_path, database):
    '''
    Function to add the IRTF Spectral Library to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    data_file = [os.path.join(input_path, 'M_fits_091201.tar'),
                 os.path.join(input_path, 'L_fits_091201.tar'),
                 os.path.join(input_path, 'T_fits_091201.tar')]

    data_folder = [os.path.join(input_path, 'M_fits_091201'),
                   os.path.join(input_path, 'L_fits_091201'),
                   os.path.join(input_path, 'T_fits_091201')]

    data_type = ['M dwarfs (7.5 MB)',
                 'L dwarfs (850 kB)',
                 'T dwarfs (100 kB)']

    url_root = 'http://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/Data/'

    url = [url_root+'M_fits_091201.tar',
           url_root+'L_fits_091201.tar',
           url_root+'T_fits_091201.tar']

    for i, item in enumerate(data_file):
        if not os.path.isfile(item):
            sys.stdout.write('Downloading IRTF Spectral Library - '+data_type[i]+'...')
            sys.stdout.flush()

            urlretrieve(url[i], item)

            sys.stdout.write(' [DONE]\n')
            sys.stdout.flush()

    sys.stdout.write('Unpacking IRTF Spectral Library...')
    sys.stdout.flush()

    for i, item in enumerate(data_file):
        tar = tarfile.open(item)
        tar.extractall(path=data_folder[i])
        tar.close()

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()

    database.create_group('spectra/irtf')

    for i, item in enumerate(data_folder):
        for root, _, files in os.walk(item):

            for _, filename in enumerate(files):
                if filename[-9:] != '_ext.fits':
                    fitsfile = os.path.join(root, filename)
                    spdata, header = fits.getdata(fitsfile, header=True)

                    name = header['OBJECT']
                    sptype = header['SPTYPE']

                    if name[-2:] == 'AB':
                        name = name[:-2]
                    elif name[-3:] == 'ABC':
                        name = name[:-3]

                    sys.stdout.write('\rAdding IRTF Spectral Library... '+'{:<40}'.format(name))
                    sys.stdout.flush()

                    simbad_id, distance = queries.get_distance(name) # [pc]

                    sptype = util.update_sptype(np.array([sptype]))[0]

                    dset = database.create_dataset('spectra/irtf/'+name,
                                                   data=spdata,
                                                   dtype='f')

                    dset.attrs['name'] = str(name)
                    dset.attrs['sptype'] = str(sptype)
                    dset.attrs['simbad'] = str(simbad_id)
                    dset.attrs['distance'] = distance

    sys.stdout.write('\rAdding IRTF Spectral Library... '+'{:<40}'.format('[DONE]')+'\n')
    sys.stdout.flush()

    database.close()



'''
Database module.
'''

import os
import sys
import warnings
import configparser

import h5py
import emcee
import numpy as np

from astropy.io import votable

from species.analysis import photometry
from species.core import box, constants
from species.read import read_model, read_calibration
from species.data import drift_phoenix, btnextgen, vega, irtf, spex, vlm_plx, leggett, \
                         companions, filters, util


warnings.simplefilter('ignore', UserWarning)


class Database:
    '''
    Text.
    '''

    def __init__(self):
        '''
        :return: None
        '''

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']
        self.input_path = config['species']['input']

    def list_items(self):
        '''
        return: None
        '''

        sys.stdout.write('Database content:\n')

        def descend(h5_object,
                    seperator=''):
            '''
            :param h5_object:
            :type h5_object: h5py._hl.files.File, h5py._hl.group.Group, h5py._hl.dataset.Dataset
            :param separator:
            :type separator: str

            :return: None
            '''

            if isinstance(h5_object, (h5py._hl.files.File, h5py._hl.group.Group)):
                for key in h5_object.keys():
                    sys.stdout.write(seperator+'- '+key+': '+str(h5_object[key])+'\n')
                    descend(h5_object[key], seperator=seperator+'\t')

            elif isinstance(h5_object, h5py._hl.dataset.Dataset):
                for key in h5_object.attrs.keys():
                    sys.stdout.write(seperator+'- '+key+': '+str(h5_object.attrs[key])+'\n')

        h5_file = h5py.File(self.database, 'r')
        descend(h5_file)
        h5_file.close()

        sys.stdout.flush()

    def list_companions(self):
        '''
        :return: None
        '''

        comp_phot = companions.get_data()

        sys.stdout.write('Database: '+self.database+'\n')
        sys.stdout.write('Directly imaged companions: ')
        sys.stdout.write(str(list(comp_phot.keys()))+'\n')
        sys.stdout.flush()

    def add_companion(self,
                      name=None):
        '''
        :param name: Companion name. All companions are added if set to None.
        :type name: tuple(str, )

        :return: None
        '''

        if isinstance(name, str):
            name = tuple((name, ))

        data = companions.get_data()

        if name is None:
            name = data.keys()

        for item in name:
            self.add_object(object_name=item,
                            distance=data[item]['distance'],
                            app_mag=data[item]['app_mag'])

    def add_filter(self,
                   filter_id,
                   filename=None):
        '''
        :param filter_id: Filter ID from the SVO Filter Profile Service (e.g., 'Paranal/NACO.Lp').
        :type filter_id: str
        :param filename: Filename with the filter profile. The first column should contain the
                         wavelength (micron) and the second column the transmission (no units).
                         The profile is downloaded from the SVO Filter Profile Service if set to
                         None.
        :type filename: str

        :return: None
        '''

        filter_split = filter_id.split('/')

        h5_file = h5py.File(self.database, 'a')

        if 'filters' not in h5_file:
            h5_file.create_group('filters')

        if 'filters/'+filter_split[0] not in h5_file:
            h5_file.create_group('filters/'+filter_split[0])

        if 'filters/'+filter_id in h5_file:
            del h5_file['filters/'+filter_id]

        sys.stdout.write('Adding filter: '+filter_id+'...')
        sys.stdout.flush()

        if filename:
            data = np.loadtxt(filename)
            wavelength = data[:, 0]
            transmission = data[:, 1]

        else:
            wavelength, transmission = filters.download_filter(filter_id)

        h5_file.create_dataset('filters/'+filter_id,
                               data=np.vstack((wavelength, transmission)),
                               dtype='f')

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

        h5_file.close()

    def add_model(self,
                  model,
                  wavelength=None,
                  teff=None):
        '''
        :param model: Model name.
        :type model: str
        :param wavelength: Wavelength (micron) range.
        :type wavelength: tuple(float, float)
        :param teff: Effective temperature (K) range.
        :type teff: tuple(float, float)

        :return: None
        '''

        h5_file = h5py.File(self.database, 'a')

        if 'models' not in h5_file:
            h5_file.create_group('models')

        if model[0:13] == 'drift-phoenix':
            drift_phoenix.add_drift_phoenix(self.input_path, h5_file)

        elif model[0:10] == 'bt-nextgen':
            btnextgen.add_btnextgen(self.input_path, h5_file, wavelength, teff)

        util.add_missing(model, h5_file)

        h5_file.close()

    def add_object(self,
                   object_name,
                   distance,
                   app_mag):
        '''
        :param object_name: Object name.
        :type object_name: str
        :param distance: Distance (pc).
        :type distance: float
        :param app_mag: Apparent magnitudes.
        :type app_mag: dict

        :return: None
        '''

        flux = {}
        error = {}

        for item in app_mag:
            synphot = photometry.SyntheticPhotometry(item)
            flux[item], error[item] = synphot.magnitude_to_flux(app_mag[item][0], app_mag[item][1])

        sys.stdout.write('Adding object: '+object_name+'...')
        sys.stdout.flush()

        h5_file = h5py.File(self.database, 'a')

        if 'objects' not in h5_file:
            h5_file.create_group('objects')

        if 'objects/'+object_name not in h5_file:
            h5_file.create_group('objects/'+object_name)

        if 'objects/'+object_name+'/distance' in h5_file:
            del h5_file['objects/'+object_name+'/distance']

        h5_file.create_dataset('objects/'+object_name+'/distance',
                               data=distance,
                               dtype='f') # [pc]

        for item in app_mag:
            if 'objects/'+object_name+'/'+item in h5_file:
                del h5_file['objects/'+object_name+'/'+item]

            data = np.asarray([app_mag[item][0],
                               app_mag[item][1],
                               flux[item],
                               error[item]])

            # [mag], [mag], [W m-2 micron-1], [W m-2 micron-1]
            h5_file.create_dataset('objects/'+object_name+'/'+item,
                                   data=data,
                                   dtype='f')

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

        h5_file.close()

    def add_photometry(self,
                       library):
        '''
        :param library: Photometry library.
        :type library: str

        :return: None
        '''

        h5_file = h5py.File(self.database, 'a')

        if 'photometry' not in h5_file:
            h5_file.create_group('photometry')

        if 'photometry/'+library in h5_file:
            del h5_file['photometry/'+library]

        if library[0:7] == 'vlm-plx':
            vlm_plx.add_vlm_plx(self.input_path, h5_file)

        elif library[0:7] == 'leggett':
            leggett.add_leggett(self.input_path, h5_file)

        h5_file.close()

    def add_calibration(self,
                        filename,
                        tag):
        '''
        Function for adding a calibration spectrum to the database.

        :param filename: Filename with the calibration spectrum. The first column should contain
                         the wavelength (micron), the second column the flux density (W m-2
                         micron-1), and the third column the error (W m-2 micron-1).
        :type filename: str
        :param tag: Tag name in the database.
        :type tag: str

        :return: None
        '''

        h5_file = h5py.File(self.database, 'a')

        if 'spectra/calibration' not in h5_file:
            h5_file.create_group('spectra/calibration')

        if 'spectra/calibration/'+tag in h5_file:
            del h5_file['spectra/calibration/'+tag]

        data = np.loadtxt(filename)

        wavelength = data[:, 0] # [micron]
        flux = data[:, 1] # [W m-2 micron-1]
        error = data[:, 2] # [W m-2 micron-1]

        sys.stdout.write('Adding calibration spectrum: '+tag+'...')
        sys.stdout.flush()

        h5_file.create_dataset('spectra/calibration/'+tag,
                               data=np.vstack((wavelength, flux, error)),
                               dtype='f')

        h5_file.close()

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    def add_spectrum(self,
                     spectrum):
        '''
        :param spectrum: Spectral library.
        :type spectrum: str

        :return: None
        '''

        h5_file = h5py.File(self.database, 'a')

        if 'spectra' not in h5_file:
            h5_file.create_group('spectra')

        if 'spectra/'+spectrum in h5_file:
            del h5_file['spectra/'+spectrum]

        if spectrum[0:5] == 'vega':
            vega.add_vega(self.input_path, h5_file)

        elif spectrum[0:5] == 'irtf':
            irtf.add_irtf(self.input_path, h5_file)

        elif spectrum[0:5] == 'spex':
            spex.add_spex(self.input_path, h5_file)

        h5_file.close()

    def add_votable(self,
                    object_name,
                    distance,
                    filename):
        '''
        :param object_name: Object name.
        :type object_name: str
        :param distance: Distance (pc).
        :type distance: float
        :param filename: Filename.
        :type filename: str

        :return: None
        '''

        # flux = {}
        # error = {}
        #
        # for item in app_mag:
        #     mag = app_mag[item]
        #
        #     synphot = photometry.SyntheticPhotometry(item)
        #     flux[item], error[item] = synphot.magnitude_to_flux(mag[0], mag[1])

        sys.stdout.write('Adding object from VOTable: '+object_name+'...')
        sys.stdout.flush()

        table = votable.parse_single_table(filename)

        frequency = table.array['sed_freq'] # Mean frequency [GHz]
        flux = table.array['_sed_flux'] # Flux density [Jy]
        error = table.array['_sed_eflux'] # [Jy]
        filter_name = table.array['_sed_filter']

        wavelength = 1e6*constants.LIGHT/(frequency*1e9) # [micron]

        conversion = 1e-6*1e-26*constants.LIGHT/(wavelength*1e-6)**2

        flux *= conversion # [W m-2 micron-1]
        error *= conversion # [W m-2 micron-1]

        h5_file = h5py.File(self.database, 'a')

        if 'objects' not in h5_file:
            h5_file.create_group('objects')

        if 'objects/'+object_name not in h5_file:
            h5_file.create_group('objects/'+object_name)

        if 'objects/'+object_name+'/distance' in h5_file:
            del h5_file['objects/'+object_name+'/distance']

        h5_file.create_dataset('objects/'+object_name+'/distance',
                               data=distance,
                               dtype='f') # [pc]

        for i, item in enumerate(filter_name):
            filter_id = util.update_filter(item)

            if filter_id is None:
                continue

            if 'objects/'+object_name+'/'+filter_id in h5_file:
                del h5_file['objects/'+object_name+'/'+filter_id]

            data = np.asarray([np.nan,
                               np.nan,
                               flux[i],
                               error[i]])

            # [mag], [mag], [W m-2 micron-1], [W m-2 micron-1]
            h5_file.create_dataset('objects/'+object_name+'/'+filter_id,
                                   data=data,
                                   dtype='f')

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

        h5_file.close()

    def add_samples(self,
                    sampler,
                    spectrum,
                    tag,
                    chisquare,
                    modelpar,
                    distance=None):
        '''
        :param sampler: Ensemble sampler.
        :type sampler: emcee.ensemble.EnsembleSampler
        :param spectrum: Tuple with the spectrum type ('model' or 'calibration') and spectrum name
                         (e.g. 'drift-phoenix').
        :type spectrum: tuple(str, str)
        :param tag: Database tag.
        :type tag: str
        :param chisquare: Maximum likelihood solution. Tuple with the chi-square value and related
                          parameter values.
        :type chisquare: tuple(float, float)
        :param modelpar: List with the model parameter names.
        :type modelpar: list(str, )
        :param distance: Distance to the object (pc). Not used if set to None.
        :type distance: float

        :return: None
        '''

        h5_file = h5py.File(self.database, 'a')

        if 'results' not in h5_file:
            h5_file.create_group('results')

        if 'results/mcmc' not in h5_file:
            h5_file.create_group('results/mcmc')

        if 'results/mcmc/'+tag in h5_file:
            del h5_file['results/mcmc/'+tag]

        dset = h5_file.create_dataset('results/mcmc/'+tag,
                                      data=sampler.chain,
                                      dtype='f')

        dset.attrs['type'] = str(spectrum[0])
        dset.attrs['spectrum'] = str(spectrum[1])
        dset.attrs['nparam'] = int(len(modelpar))

        if distance:
            dset.attrs['distance'] = float(distance)

        for i, item in enumerate(modelpar):
            dset.attrs['parameter'+str(i)] = str(item)

        dset.attrs['min_chi'] = float(chisquare[0])

        for i, item in enumerate(modelpar):
            dset.attrs['chisquare'+str(i)] = float(chisquare[1][item])

        mean_accep = np.mean(sampler.acceptance_fraction)
        dset.attrs['acceptance'] = float(mean_accep)

        sys.stdout.write('Mean acceptance fraction: {0:.3f}'.format(mean_accep)+'\n')
        sys.stdout.flush()

        try:
            int_auto = emcee.autocorr.integrated_time(sampler.flatchain)

            sys.stdout.write('Integrated autocorrelation time = '+str(int_auto)+'\n')
            sys.stdout.flush()

        except emcee.autocorr.AutocorrError:
            int_auto = None

        if int_auto is not None:
            for i, item in enumerate(int_auto):
                dset.attrs['autocorrelation'+str(i)] = float(item)

        h5_file.close()

    def get_chisquare(self,
                      tag):
        '''
        :param tag:
        :type tag: str

        :return:
        :rtype: species.core.box.SamplesBox
        '''

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file['results/chisquare/'+tag]

        nparam = dset.attrs['nparam']
        spectrum = dset.attrs['spectrum']

        param = {}
        for i in range(nparam):
            par_key = dset.attrs['parameter'+str(i)]
            par_value = dset.attrs[par_key]

            param[par_key] = par_value

        h5_file.close()

        return spectrum, param

    def get_mcmc_spectra(self,
                         tag,
                         burnin,
                         random,
                         wavelength,
                         sampling=None):
        '''
        :param tag: Database tag with the MCMC samples.
        :type tag: str
        :param burnin: Number of burnin steps.
        :type burnin: int
        :param random: Number of random samples.
        :type random: int
        :param wavelength: Wavelength range (micron) or filter name. Full spectrum if set to None.
        :type wavelength: tuple(float, float) or str
        :param sampling: Spectral sampling.
        :type sampling: tuple

        :return: Boxes with the randomly sampled spectra.
        :rtype: tuple(species.core.box.ModelBox, )
        '''

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file['results/mcmc/'+tag]

        nparam = dset.attrs['nparam']
        spectrum_type = dset.attrs['type']
        spectrum_name = dset.attrs['spectrum']

        if dset.attrs.__contains__('distance'):
            distance = dset.attrs['distance']
        else:
            distance = None

        samples = np.asarray(dset)
        samples = samples[:, burnin:, :]

        ran_walker = np.random.randint(samples.shape[0], size=random)
        ran_step = np.random.randint(samples.shape[1], size=random)
        samples = samples[ran_walker, ran_step, :]

        param = []
        for i in range(nparam):
            param.append(str(dset.attrs['parameter'+str(i)]))

        if spectrum_type == 'model':
            readmodel = read_model.ReadModel(spectrum_name, wavelength)
        elif spectrum_type == 'calibration':
            readcalib = read_calibration.ReadCalibration(spectrum_name, None)

        boxes = []

        for i in range(samples.shape[0]):
            model_par = {}
            for j in range(samples.shape[1]):
                model_par[param[j]] = samples[i, j]

            if distance:
                model_par['distance'] = distance

            if spectrum_type == 'model':
                specbox = readmodel.get_model(model_par, sampling)
            elif spectrum_type == 'calibration':
                specbox = readcalib.get_spectrum(model_par)

            box.type = 'mcmc'

            boxes.append(specbox)

        h5_file.close()

        return tuple(boxes)

    def get_object(self,
                   object_name,
                   filter_id):
        '''
        :param object_name: Object name in the database.
        :type object_name: str
        :param filter_id: Filter IDs for which the photometry is selected. All available photometry
                          of the object is selected if set to None.
        :type filter_id: tuple(str, )

        :return: Box with the object.
        :rtype: species.core.box.ObjectBox
        '''

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file['objects/'+object_name]

        distance = np.asarray(dset['distance'])

        magnitude = {}
        flux = {}

        if filter_id:
            for item in filter_id:
                data = dset[item]

                magnitude[item] = np.asarray(data[0:2])
                flux[item] = np.asarray(data[2:4])

        else:
            for key in dset.keys():
                if key != 'distance':
                    for item in dset[key]:
                        name = key+'/'+item

                        magnitude[name] = np.asarray(dset[name][0:2])
                        flux[name] = np.asarray(dset[name][2:4])

        h5_file.close()

        return box.create_box('object',
                              name=object_name,
                              filter=tuple(magnitude.keys()),
                              magnitude=magnitude,
                              flux=flux,
                              distance=distance)

    def get_samples(self,
                    tag,
                    burnin=None,
                    random=None):
        '''
        :param tag:
        :type tag: str
        :param burnin:
        :type burnin: int
        :param random:
        :type random: int

        :return:
        :rtype: species.core.box.SamplesBox
        '''

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file['results/mcmc/'+tag]

        spectrum = dset.attrs['spectrum']
        nparam = dset.attrs['nparam']

        samples = np.asarray(dset)

        if burnin:
            samples = samples[:, burnin:, :]

        if random:
            ran_walker = np.random.randint(samples.shape[0], size=random)
            ran_step = np.random.randint(samples.shape[1], size=random)
            samples = samples[ran_walker, ran_step, :]

        param = []
        chisquare = []
        for i in range(nparam):
            param.append(dset.attrs['parameter'+str(i)])
            chisquare.append(dset.attrs['chisquare'+str(i)])

        h5_file.close()

        return box.create_box('samples',
                              spectrum=spectrum,
                              parameters=param,
                              samples=samples,
                              chisquare=chisquare)



'''
Setup module.
'''

import os
import sys
import configparser

import h5py

import species


class SpeciesInit:
    '''
    Text.
    '''

    def __init__(self,
                 config_path):
        '''
        :param config_path: Location of the configuration file (species_config.ini).
        :type config_path: str

        :return: None
        '''

        sys.stdout.write('Initiating species v'+species.__version__+'...')
        sys.stdout.flush()

        self.config_path = config_path

        config_file = os.path.join(self.config_path, 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        database_file = config['species']['database']
        input_path = config['species']['input']

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

        if not os.path.exists(input_path):
            sys.stdout.write('Creating input folder...')
            sys.stdout.flush()

            os.makedirs(input_path)

            sys.stdout.write(' [DONE]\n')
            sys.stdout.flush()

        if not os.path.isfile(database_file):
            sys.stdout.write('Creating species_database.hdf5...')
            sys.stdout.flush()

            h5_file = h5py.File(database_file, 'w')
            h5_file.close()

            sys.stdout.write(' [DONE]\n')
            sys.stdout.flush()

        if not os.path.isfile(config_file):
            sys.stdout.write('Creating species_config.ini...')
            sys.stdout.flush()

            config = configparser.ConfigParser()

            config['species'] = {'database': database_file,
                                 'config':config_file,
                                 'input':input_path}

            with open(config_file, 'w') as config_ini:
                config.write(config_ini)

            sys.stdout.write(' [DONE]\n')
            sys.stdout.flush()



'''
Text
'''

import sys
import math

import emcee
import progress.bar
import numpy as np

from species.analysis import photometry
from species.data import database
from species.read import read_model, read_object


MIN_CHISQ = np.inf
MIN_PARAM = None


def lnprior(param,
            bounds,
            modelpar,
            prior):
    '''
    :param param: Parameter values.
    :type param: numpy.ndarray
    :param bounds: Parameter boundaries.
    :type bounds: dict
    :param modelpar: Parameter names.
    :type modelpar: tuple(str, )
    :param prior: Gaussian prior on one of the parameters. Currently only possible for the mass,
                  e.g. ('mass', 13., 3.) for an expected mass of 13 Mjup with an uncertainty of
                  3 Mjup. Not used if set to None.
    :type prior: tuple(str, float, float)

    :return: Log prior probability.
    :rtype: float
    '''

    if prior:

        modeldict = {}
        for i, item in enumerate(modelpar):
            modeldict[item] = param[i]

    for i, item in enumerate(modelpar):

        if bounds[item][0] <= param[i] <= bounds[item][1]:

            if prior is None:
                ln_prior = 0.

            elif prior[0] == 'mass':
                mass = read_model.get_mass(modeldict)
                ln_prior = -0.5*(mass-prior[1])**2/prior[2]**2

        else:
            ln_prior = -np.inf
            break

    return ln_prior


def lnlike(param,
           modelpar,
           modelphot,
           objphot,
           synphot,
           sampling,
           distance):
    '''
    :param param:
    :type param:
    :param modelpar:
    :type modelpar:
    :param modelphot:
    :type modelphot:
    :param objphot:
    :type objphot:
    :param synphot:
    :type synphot:
    :param sampling:
    :type sampling:
    :param distance:
    :type distance:

    :return: Log likelihood probability.
    :rtype: float
    '''

    global MIN_CHISQ
    global MIN_PARAM

    paramdict = {}
    for i, item in enumerate(modelpar):
        paramdict[item] = param[i]

    paramdict['distance'] = distance

    chisq = 0.

    for i, item in enumerate(objphot):
        flux = modelphot[i].get_photometry(paramdict, sampling, synphot[i])
        chisq += (item[0]-flux)**2 / item[1]**2

    if chisq < MIN_CHISQ:
        MIN_CHISQ = chisq
        MIN_PARAM = paramdict

    return -0.5*chisq


def lnprob(param,
           bounds,
           modelpar,
           modelphot,
           objphot,
           synphot,
           sampling,
           distance,
           prior):
    '''
    :param param:
    :type param:
    :param bounds:
    :type bounds:
    :param modelpar:
    :type modelpar:
    :param modelphot:
    :type modelphot:
    :param objphot:
    :type objphot:
    :param synphot:
    :type synphot:
    :param sampling:
    :type sampling:
    :param distance:
    :type distance:
    :param prior: Gaussian prior. Not used if set to None.
    :type prior: tuple(str, float, float)

    :return:
    :rtype:
    '''

    ln_prior = lnprior(param, bounds, modelpar, prior)

    if math.isinf(ln_prior):
        ln_prob = -np.inf

    else:
        ln_prob = ln_prior + lnlike(param,
                                    modelpar,
                                    modelphot,
                                    objphot,
                                    synphot,
                                    sampling,
                                    distance)

    return ln_prob


class FitModel:
    '''
    Text
    '''

    def __init__(self,
                 objname,
                 filters,
                 model,
                 sampling,
                 bounds):
        '''
        :param objname: Object name in the database.
        :type objname: str
        :param filters: Filter IDs for which the photometry is selected. All available
                        photometry of the object is selected if set to None.
        :type filters: tuple(str, )
        :name model: Atmospheric model.
        :type model: str
        :name sampling: Wavelength sampling for the computation of synthetic photometry
                        ('specres' or 'gaussian').
        :type sampling: tuple
        :name bounds: Parameter boundaries. Full parameter range is used if None or not specified.
                      The radius parameter range is set to 0-5 Rjup if not specified.
        :type bounds: dict

        :return: None
        '''

        self.object = read_object.ReadObject(objname)
        self.distance = self.object.get_distance()

        self.model = model
        self.sampling = sampling
        self.bounds = bounds

        self.objphot = []
        self.modelphot = []
        self.synphot = []

        if self.bounds and 'teff' in self.bounds:
            teff_bound = self.bounds['teff']
        else:
            teff_bound = None

        if self.bounds:
            readmodel = read_model.ReadModel(self.model, None, teff_bound)
            bounds_grid = readmodel.get_bounds()

            for item in bounds_grid:
                if item not in self.bounds:
                    self.bounds[item] = bounds_grid[item]

        else:
            readmodel = read_model.ReadModel(self.model, None, None)
            self.bounds = readmodel.get_bounds()

        if 'radius' not in self.bounds:
            self.bounds['radius'] = (0., 5.)

        if filters is None:
            species_db = database.Database()
            objectbox = species_db.get_object(objname, None)
            filters = objectbox.filter

        for item in filters:
            readmodel = read_model.ReadModel(self.model, item, teff_bound)
            readmodel.interpolate()
            self.modelphot.append(readmodel)

            sphot = photometry.SyntheticPhotometry(item)
            self.synphot.append(sphot)

            obj_phot = self.object.get_photometry(item)
            self.objphot.append((obj_phot[2], obj_phot[3]))

        self.modelpar = readmodel.get_parameters()
        self.modelpar.append('radius')

    def run_mcmc(self,
                 nwalkers,
                 nsteps,
                 guess,
                 tag,
                 prior=None,
                 ncpu=1):
        '''
        :return: None
        '''

        global MIN_CHISQ
        global MIN_PARAM

        sigma = {'teff':5., 'logg':0.01, 'feh':0.01, 'radius':0.01}

        sys.stdout.write('Running MCMC...')
        sys.stdout.flush()

        ndim = len(self.bounds)

        initial = np.zeros((nwalkers, ndim))
        for i, item in enumerate(self.modelpar):
            if guess[item] is not None:
                initial[:, i] = guess[item] + np.random.normal(0, sigma[item], nwalkers)

            else:
                initial[:, i] = np.random.uniform(low=self.bounds[item][0],
                                                  high=self.bounds[item][1],
                                                  size=nwalkers)

        sampler = emcee.EnsembleSampler(nwalkers=nwalkers,
                                        dim=ndim,
                                        lnpostfn=lnprob,
                                        a=2.,
                                        args=([self.bounds,
                                               self.modelpar,
                                               self.modelphot,
                                               self.objphot,
                                               self.synphot,
                                               self.sampling,
                                               self.distance,
                                               prior]),
                                        threads=ncpu)

        progbar = progress.bar.Bar('\rRunning MCMC...',
                                   max=nsteps,
                                   suffix='%(percent)d%%')

        for i, _ in enumerate(sampler.sample(initial, iterations=nsteps)):
            progbar.next()

        progbar.finish()

        species_db = database.Database()

        species_db.add_samples(sampler=sampler,
                               spectrum=('model', self.model),
                               tag=tag,
                               chisquare=(MIN_CHISQ, MIN_PARAM),
                               modelpar=self.modelpar,
                               distance=self.distance)



'''
Text
'''

import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

import numpy as np


requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def download_filter(filter_id):
    '''
    :param filter_id: Filter ID.
    :type filter_id: str

    :return:
    :rtype: numpy.ndarray, numpy.ndarray
    '''

    if filter_id == 'LCO/VisAO.Ys':
        url = 'https://visao.as.arizona.edu/software_files/visao/html/VisAO_Ys_filter_curve.dat'

        session = requests.Session()
        response = session.get(url, verify=False)
        data = response.content

        wavelength = []
        transmission = []
        for line in data.splitlines():
            if not line.startswith(b'#'):
                split = line.split()

                wavelength.append(float(split[0])) # [micron]
                transmission.append(float(split[1]))

        wavelength = np.asarray(wavelength)
        transmission = np.asarray(transmission)

        wavelength = wavelength[:-7]
        transmission = transmission[:-7]

    else:
        url = 'http://svo2.cab.inta-csic.es/svo/theory/fps/getdata.php?format=ascii&id='+filter_id

        session = requests.Session()
        response = session.get(url)
        data = response.content

        wavelength = []
        transmission = []
        for line in data.splitlines():
            if not line.startswith(b'#'):
                split = line.split(b' ')

                wavelength.append(float(split[0])*1e-4) # [micron]
                transmission.append(float(split[1]))

        wavelength = np.asarray(wavelength)
        transmission = np.asarray(transmission)

    return wavelength, transmission



'''
Module with functions for making plots.
'''

import os
import sys
import math
import itertools

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from species.core import box
from species.plot import util
from species.read import read_filter


mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
mpl.rcParams['font.family'] = 'serif'

plt.rc('axes', edgecolor='black', linewidth=2)


def plot_spectrum(boxes,
                  filters,
                  output,
                  colors=None,
                  residuals=None,
                  xlim=None,
                  ylim=None,
                  scale=('linear', 'linear'),
                  title=None,
                  offset=None):
    '''
    :param boxes:
    :type boxes: tuple(species.analysis.box.SpectrumBox and/or
                 species.analysis.box.PhotometryBox and/or
                 species.analysis.box.ModelBox)
    :param filters:
    :type filters: tuple(str, )
    :param output:
    :type output: str

    :return: None
    '''

    marker = itertools.cycle(('o', 's', '*', 'p', '<', '>', 'P', 'v', '^'))

    if residuals and filters:
        plt.figure(1, figsize=(7, 5))
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[1, 3, 1])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[1, 0])
        ax2 = plt.subplot(gridsp[0, 0])
        ax3 = plt.subplot(gridsp[2, 0])

    elif residuals:
        plt.figure(1, figsize=(7, 5))
        gridsp = mpl.gridspec.GridSpec(2, 1, height_ratios=[4, 1])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])
        ax3 = plt.subplot(gridsp[1, 0])

    elif filters:
        plt.figure(1, figsize=(7, 5))
        gridsp = mpl.gridspec.GridSpec(2, 1, height_ratios=[1, 4])
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[1, 0])
        ax2 = plt.subplot(gridsp[0, 0])

    else:
        plt.figure(1, figsize=(7, 4))
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])

    ax1.grid(True, linestyle=':', linewidth=0.7, color='gray', dashes=(1, 4), alpha=0.3, zorder=0)

    if residuals:
        labelbottom = False
    else:
        labelbottom = True

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True, labelbottom=labelbottom)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=0.8, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True, labelbottom=labelbottom)

    if filters:
        ax2.grid(True, linestyle=':', linewidth=0.7, color='gray', dashes=(1, 4),
                 alpha=0.3, zorder=0)

        ax2.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                        direction='in', width=0.8, length=5, labelsize=12, top=True,
                        bottom=True, left=True, right=True, labelbottom=False)

        ax2.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                        direction='in', width=0.8, length=3, labelsize=12, top=True,
                        bottom=True, left=True, right=True, labelbottom=False)

    if residuals:
        ax3.grid(True, linestyle=':', linewidth=0.7, color='gray', dashes=(1, 4),
                 alpha=0.3, zorder=0)

        ax3.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                        direction='in', width=0.8, length=5, labelsize=12, top=True,
                        bottom=True, left=True, right=True)

        ax3.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                        direction='in', width=0.8, length=3, labelsize=12, top=True,
                        bottom=True, left=True, right=True)

    if residuals and filters:
        ax1.set_xlabel('', fontsize=13)
        ax2.set_xlabel('', fontsize=13)
        ax3.set_xlabel('Wavelength [micron]', fontsize=13)

    elif residuals:
        ax1.set_xlabel('', fontsize=13)
        ax3.set_xlabel('Wavelength [micron]', fontsize=13)

    elif filters:
        ax1.set_xlabel('Wavelength [micron]', fontsize=13)
        ax2.set_xlabel('', fontsize=13)

    else:
        ax1.set_xlabel('Wavelength [micron]', fontsize=13)

    if filters:
        ax2.set_ylabel('Transmission', fontsize=13)

    if residuals:
        ax3.set_ylabel('Residual [$\sigma$]', fontsize=13)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])
    else:
        ax1.set_xlim(0.6, 6.)

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

        ylim = ax1.get_ylim()

        exponent = math.floor(math.log10(ylim[1]))
        scaling = 10.**exponent

        ax1.set_ylabel('Flux [10$^{'+str(exponent)+'}$ W m$^{-2}$ $\mu$m$^{-1}$]', fontsize=13)
        ax1.set_ylim(ylim[0]/scaling, ylim[1]/scaling)

        if ylim[0] < 0.:
            ax1.axhline(0.0, linestyle='--', color='gray', dashes=(2, 4), zorder=1)

    else:
        ax1.set_ylabel('Flux [W m$^{-2}$ $\mu$m$^{-1}$]', fontsize=13)
        scaling = 1.

    if filters:
        ax2.set_ylim(0., 1.)

    xlim = ax1.get_xlim()

    if filters:
        ax2.set_xlim(xlim[0], xlim[1])

    if residuals:
        ax3.set_xlim(xlim[0], xlim[1])

    if offset and residuals and filters:
        ax3.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax2.get_yaxis().set_label_coords(offset[1], 0.5)
        ax3.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset and filters:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax2.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset and residuals:
        ax3.get_xaxis().set_label_coords(0.5, offset[0])

        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
        ax3.get_yaxis().set_label_coords(offset[1], 0.5)

    elif offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.12)
        ax1.get_yaxis().set_label_coords(-0.1, 0.5)

    ax1.set_xscale(scale[0])
    ax1.set_yscale(scale[1])

    if filters:
        ax2.set_xscale(scale[0])

    if residuals:
        ax3.set_xscale(scale[0])

    for j, boxitem in enumerate(boxes):

        if isinstance(boxitem, (box.SpectrumBox, box.ModelBox)):
            wavelength = boxitem.wavelength
            flux = boxitem.flux

            if isinstance(wavelength[0], (np.float32, np.float64)):
                data = np.array(flux, dtype=np.float64)
                masked = np.ma.array(data, mask=np.isnan(data))

                if isinstance(boxitem, box.ModelBox):
                    param = boxitem.parameters

                    par_key, par_unit = util.quantity_unit(list(param.keys()))
                    par_val = list(param.values())

                    label = ''
                    for i, item in enumerate(par_key):

                        if item == '$T_\mathregular{eff}$':
                            value = '{:.1f}'.format(par_val[i])
                        elif item in ('$\log\,g$', '$R$', '$M$', '[Fe/H]'):
                            value = '{:.2f}'.format(par_val[i])
                        elif item == '$L$':
                            value = '{0:.1e}'.format(par_val[i])
                        else:
                            continue

                        label += item+' = '+str(value)+' '+par_unit[i]

                        if i < len(par_key)-1:
                            label += ', '

                else:
                    label = None

                if colors:
                    ax1.plot(wavelength, masked/scaling, color=colors[j], lw=1.0, label=label, zorder=4)
                else:
                    ax1.plot(wavelength, masked/scaling, lw=1.0, label=label, zorder=4)

            elif isinstance(wavelength[0], (np.ndarray)):
                for i, item in enumerate(wavelength):
                    data = np.array(flux[i], dtype=np.float64)
                    masked = np.ma.array(data, mask=np.isnan(data))

                    ax1.plot(item, masked/scaling, lw=1)

        elif isinstance(boxitem, tuple):
            for i, item in enumerate(boxitem):
                wavelength = item.wavelength
                flux = item.flux

                data = np.array(flux, dtype=np.float64)
                masked = np.ma.array(data, mask=np.isnan(data))

                if colors:
                    ax1.plot(wavelength, masked/scaling, lw=0.4, color=colors[j], alpha=0.5, zorder=3)
                else:
                    ax1.plot(wavelength, masked/scaling, lw=0.4, alpha=0.5, zorder=3)

        elif isinstance(boxitem, box.PhotometryBox):
            if colors:
                ax1.plot(boxitem.wavelength, boxitem.flux/scaling, marker=next(marker), ms=6, \
                         color=colors[j], label=boxitem.name, zorder=5)
            else:
                ax1.plot(boxitem.wavelength, boxitem.flux/scaling, marker=next(marker), ms=6, \
                         label=boxitem.name, zorder=5)

        elif isinstance(boxitem, box.ObjectBox):
            for item in boxitem.flux:
                transmission = read_filter.ReadFilter(item)
                wavelength = transmission.mean_wavelength()
                fwhm = transmission.filter_fwhm()

                ax1.errorbar(wavelength, boxitem.flux[item][0]/scaling, xerr=fwhm/2.,
                             yerr=boxitem.flux[item][1]/scaling, marker='s', ms=5, zorder=5,
                             color=colors[j], markerfacecolor=colors[j])

        elif isinstance(boxitem, box.SynphotBox):
            for item in boxitem.flux:
                transmission = read_filter.ReadFilter(item)
                wavelength = transmission.mean_wavelength()
                fwhm = transmission.filter_fwhm()

                ax1.errorbar(wavelength, boxitem.flux[item]/scaling, xerr=fwhm/2., yerr=None,
                             alpha=0.7, marker='s', ms=5, zorder=6, color=colors[j],
                             markerfacecolor='white')

    if filters:
        for i, item in enumerate(filters):
            transmission = read_filter.ReadFilter(item)
            data = transmission.get_filter()

            ax2.plot(data[0, ], data[1, ], '-', lw=0.7, color='black')

    if residuals:
        diff = np.zeros((2, len(residuals[0].flux)))

        for i, item in enumerate(residuals[0].flux):
            transmission = read_filter.ReadFilter(item)
            diff[0, i] = transmission.mean_wavelength()
            diff[1, i] = (residuals[0].flux[item][0]-residuals[1].flux[item]) / \
                residuals[0].flux[item][1]

        ax3.plot(diff[0, ], diff[1, ], marker='s', ms=5, linestyle='none', color='black')

        res_lim = math.ceil(np.amax(np.abs(diff[1, ])))

        ax3.axhline(0.0, linestyle='--', color='gray', dashes=(2, 4), zorder=1)
        ax3.set_ylim(-res_lim, res_lim)

    if filters:
        ax2.set_ylim(0., 1.1)

    sys.stdout.write('Plotting spectrum: '+output+'...')
    sys.stdout.flush()

    # ax1.xaxis.set_major_formatter(FormatStrFormatter('%.1f'))
    # ax1.xaxis.set_minor_formatter(FormatStrFormatter('%.1f'))

    if title:
        if filters:
            ax2.set_title(title, y=1.02, fontsize=15)
        else:
            ax1.set_title(title, y=1.02, fontsize=15)

    handles, _ = ax1.get_legend_handles_labels()

    if handles:
        ax1.legend(loc='upper left', prop={'size':9}, frameon=False)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.close()

    sys.stdout.write(' [DONE]\n')
    sys.stdout.flush()



'''
Text
'''

import sys
import math

import emcee
import progress.bar
import numpy as np

from species.analysis import photometry
from species.data import database
from species.read import read_object, read_calibration


MIN_CHISQ = np.inf
MIN_PARAM = None


def lnprob(param,
           bounds,
           objphot,
           specphot):
    '''
    :param param: Values of the offset and scaling parameter.
    :type param: numpy.ndarray
    :param bounds: Boundaries of the offset and scaling parameter
    :type bounds: dict
    :param objphot:
    :type objphot:
    :param specphot:
    :type specphot:

    :return:
    :rtype:
    '''

    global MIN_CHISQ
    global MIN_PARAM

    if bounds['offset'][0] <= param[0] <= bounds['offset'][1] and \
       bounds['scaling'][0] <= param[1] <= bounds['scaling'][1]:
        ln_prior = 0.

    else:
        ln_prior = -np.inf

    if math.isinf(ln_prior):
        ln_prob = -np.inf

    else:
        chisq = 0.
        for i, _ in enumerate(objphot):
            chisq += (objphot[i][0]-(param[0]+param[1]*specphot[i]))**2 / objphot[i][1]**2

        if chisq < MIN_CHISQ:
            MIN_CHISQ = chisq
            MIN_PARAM = {'offset':param[0], 'scaling':param[1]}

        ln_prob = ln_prior - 0.5*chisq

    return ln_prob


class FitSpectrum:
    '''
    Text
    '''

    def __init__(self,
                 objname,
                 filters,
                 spectrum,
                 bounds):
        '''
        :param objname: Object name in the database.
        :type objname: str
        :param filters: Filter IDs for which the photometry is selected. All available
                        photometry of the object is selected if set to None.
        :type filters: tuple(str, )
        :param spectrum: Calibration spectrum.
        :type spectrum: str
        :param bounds: Boundaries of the offset and scaling parameter, as
                       {'offset':(min, max), 'scaling':(min, max)}.
        :type bounds: dict

        :return: None
        '''

        self.object = read_object.ReadObject(objname)

        self.spectrum = spectrum
        self.bounds = bounds

        self.objphot = []
        self.specphot = []

        if filters is None:
            species_db = database.Database()
            objectbox = species_db.get_object(objname, None)
            filters = objectbox.filter

        for item in filters:
            readcalib = read_calibration.ReadCalibration(self.spectrum, item)
            calibspec = readcalib.get_spectrum()

            synphot = photometry.SyntheticPhotometry(item)
            spec_phot = synphot.spectrum_to_photometry(calibspec.wavelength, calibspec.flux)
            self.specphot.append(spec_phot)

            obj_phot = self.object.get_photometry(item)
            self.objphot.append((obj_phot[2], obj_phot[3]))

        self.modelpar = ['offset', 'scaling']

    def run_mcmc(self,
                 nwalkers,
                 nsteps,
                 guess,
                 tag):
        '''
        :param nwalkers: Number of walkers.
        :type nwalkers: int
        :param nsteps: Number of steps for each walker.
        :type nsteps: int
        :param guess: Guess of the offset and scaling parameter, as
                      {'offset':(guess), 'scaling':(guess)}. Non-zero values work best.
        :type guess: dict
        :param tag: Database tag for the results.
        :type tag: int

        :return: None
        '''

        global MIN_CHISQ
        global MIN_PARAM

        sys.stdout.write('Running MCMC...')
        sys.stdout.flush()

        ndim = 2

        initial = np.zeros((nwalkers, ndim))

        initial[:, 0] = guess['offset'] + np.random.normal(0, 1e-1*guess['offset'], nwalkers)
        initial[:, 1] = guess['scaling'] + np.random.normal(0, 1e-1*guess['scaling'], nwalkers)

        sampler = emcee.EnsembleSampler(nwalkers=nwalkers,
                                        dim=ndim,
                                        lnpostfn=lnprob,
                                        a=2.,
                                        args=([self.bounds,
                                               self.objphot,
                                               self.specphot]))

        progbar = progress.bar.Bar('\rRunning MCMC...',
                                   max=nsteps,
                                   suffix='%(percent)d%%')

        for _ in sampler.sample(initial, iterations=nsteps):
            progbar.next()

        progbar.finish()

        species_db = database.Database()

        species_db.add_samples(sampler=sampler,
                               spectrum=('calibration', self.spectrum),
                               tag=tag,
                               chisquare=(MIN_CHISQ, MIN_PARAM),
                               modelpar=self.modelpar,
                               distance=None)



'''
Module for BT-NextGen atmospheric models.
'''

import os
import sys
import tarfile

from urllib.request import urlretrieve

import numpy as np
import pandas as pd

from scipy.interpolate import interp1d

from species.data import util


def add_btnextgen(input_path,
                  database,
                  wl_bound,
                  teff_bound):
    '''
    Function for adding the BT-NextGen atmospheric models to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File
    :param wl_bound: Wavelength (micron) range.
    :type wl_bound: tuple(float, float)
    :param teff_bound: Effective temperature (K) range.
    :type teff_bound: tuple(float, float)

    :return: None
    '''

    spec_res = 1e5

    if not wl_bound:
        wl_bound = (1e-2, 1e2)

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'bt-nextgen/')

    files = ['BT-NextGen_M-0.0_a+0.0_hot.tar',
             'BT-NextGen_M+0.3_a+0.0_hot.tar',
             'BT-NextGen_M+0.5_a+0.0_hot.tar']

    urls = ['https://phoenix.ens-lyon.fr/Grids/BT-NextGen/SPECTRA/BT-NextGen_M-0.0_a+0.0_hot.tar',
            'https://phoenix.ens-lyon.fr/Grids/BT-NextGen/SPECTRA/BT-NextGen_M+0.3_a+0.0_hot.tar',
            'https://phoenix.ens-lyon.fr/Grids/BT-NextGen/SPECTRA/BT-NextGen_M+0.5_a+0.0_hot.tar']

    labels = ['[Fe/H]=0.0 (5.9 GB)',
              '[Fe/H]=0.3 (6.2 GB)',
              '[Fe/H]=0.5 (6.4 GB)']

    for i, item in enumerate(files):
        data_file = os.path.join(input_path, item)

        if not os.path.isfile(data_file):
            sys.stdout.write('Downloading BT-NextGen model spectra '+labels[i]+'...')
            sys.stdout.flush()

            urlretrieve(urls[i], data_file)

            sys.stdout.write(' [DONE]\n')
            sys.stdout.flush()

        sys.stdout.write('Unpacking BT-NextGen model spectra '+labels[i]+'...')
        sys.stdout.flush()

        tar = tarfile.open(data_file)
        tar.extractall(data_folder)
        tar.close()

        sys.stdout.write(' [DONE]\n')
        sys.stdout.flush()

    teff = []
    logg = []
    feh = []
    flux = []

    wavelength = [wl_bound[0]]

    while wavelength[-1] <= wl_bound[1]:
        wavelength.append(wavelength[-1] + wavelength[-1]/spec_res)

    wavelength = np.asarray(wavelength[:-1])

    for _, _, file_list in os.walk(data_folder):
        for filename in sorted(file_list):

            if filename.startswith('lte'):
                sys.stdout.write('\rAdding BT-NextGen model spectra... '+filename)
                sys.stdout.flush()

                teff_val = float(filename[3:6])*100.

                if teff_bound[0] <= teff_val <= teff_bound[1]:
                    teff.append(teff_val)
                    logg.append(float(filename[7:9]))
                    feh.append(float(filename[11:14]))

                else:
                    continue

                dataf = pd.pandas.read_csv(data_folder+filename,
                                           usecols=[0, 1],
                                           names=['wavelength', 'flux'],
                                           header=None,
                                           delim_whitespace=True,
                                           compression='bz2')

                dataf['wavelength'] = dataf['wavelength'].str.replace('D', 'E')
                dataf['flux'] = dataf['flux'].str.replace('D', 'E')

                dataf = dataf.apply(pd.to_numeric)
                data = dataf.values

                # [Angstrom] -> [micron]
                data_wavel = data[:, 0]*1e-4

                # See https://phoenix.ens-lyon.fr/Grids/FORMAT
                data_flux = 10.**(data[:, 1]-8.) # [erg s-1 cm-2 Angstrom-1]

                # [erg s-1 cm-2 Angstrom-1] -> [W m-2 micron-1]
                data_flux = data_flux*1e-7*1e4*1e4

                indices = np.where((data_wavel >= wl_bound[0]) & \
                                   (data_wavel <= wl_bound[1]))[0]

                data_wavel = data_wavel[indices]
                data_flux = data_flux[indices]

                flux_interp = interp1d(data_wavel,
                                       data_flux,
                                       kind='cubic',
                                       bounds_error=False,
                                       fill_value=float('nan'))

                flux.append(flux_interp(wavelength))

    data_sorted = util.sort_data(np.asarray(teff),
                                 np.asarray(logg),
                                 np.asarray(feh),
                                 wavelength,
                                 np.asarray(flux))

    util.write_data('bt-nextgen', database, data_sorted)

    sys.stdout.write('\rAdding BT-NextGen model spectra... [DONE]                              \n')
    sys.stdout.flush()



'''
Text
'''

import os
import sys
import warnings

from urllib.request import urlretrieve

import numpy as np

from astropy.io.votable import parse_single_table
from astropy.utils.exceptions import AstropyWarning

from species.analysis import photometry
from species.data import queries, util
from species.read import read_filter


warnings.simplefilter('ignore', category=AstropyWarning)


def add_spex(input_path, database):
    '''
    Function for adding the SpeX Prism Spectral Library to the database.

    :param input_path:
    :type input_path: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    database.create_group('spectra/spex')

    data_path = os.path.join(input_path, 'spex')

    if not os.path.exists(data_path):
        os.makedirs(data_path)

    url_all = 'http://svo2.cab.inta-csic.es/vocats/v2/spex/' \
              'cs.php?RA=180.000000&DEC=0.000000&SR=180.000000&VERB=2'

    xml_file = os.path.join(data_path, 'spex.xml')

    urlretrieve(url_all, xml_file)

    table = parse_single_table(xml_file)
    name = table.array['name']
    twomass = table.array['name2m']
    url = table.array['access_url']

    os.remove(xml_file)

    for i, item in enumerate(url):
        xml_file = os.path.join(data_path, twomass[i].decode('utf-8')+'.xml')
        urlretrieve(item.decode('utf-8'), xml_file)

        table = parse_single_table(xml_file)
        name = table.array['ID']
        name = name[0].decode('utf-8')
        url = table.array['access_url']

        sys.stdout.write('\rDownloading SpeX Prism Spectral Library... '+'{:<40}'.format(name))
        sys.stdout.flush()

        os.remove(xml_file)

        xml_file = os.path.join(data_path, name+'.xml')
        urlretrieve(url[0].decode('utf-8'), xml_file)

    sys.stdout.write('\rDownloading SpeX Prism Spectral Library... '+'{:<40}'.format('[DONE]')+'\n')
    sys.stdout.flush()

    h_twomass = photometry.SyntheticPhotometry('2MASS/2MASS.H')

    transmission = read_filter.ReadFilter('2MASS/2MASS.H')
    transmission.get_filter()

    # 2MASS H band zero point for 0 mag (Cogen et al. 2003)
    h_zp = 1.133e-9 # [W m-2 micron-1]

    for votable in os.listdir(data_path):
        if votable.endswith('.xml'):
            xml_file = os.path.join(data_path, votable)

            table = parse_single_table(xml_file)

            wavelength = table.array['wavelength'] # [Angstrom]
            flux = table.array['flux'] # Normalized units

            wavelength = np.array(wavelength*1e-4) # [micron]
            flux = np.array(flux)

            # 2MASS magnitudes
            j_mag = table.get_field_by_id('jmag').value
            h_mag = table.get_field_by_id('hmag').value
            ks_mag = table.get_field_by_id('ksmag').value

            if j_mag == b'':
                j_mag = np.nan
            else:
                j_mag = float(j_mag)

            if h_mag == b'':
                h_mag = np.nan
            else:
                h_mag = float(h_mag)

            if ks_mag == b'':
                ks_mag = np.nan
            else:
                ks_mag = float(ks_mag)

            name = table.get_field_by_id('name').value
            name = name.decode('utf-8')
            twomass_id = table.get_field_by_id('name2m').value

            sys.stdout.write('\rAdding SpeX Prism Spectral Library... '+'{:<40}'.format(name))
            sys.stdout.flush()

            try:
                sptype = table.get_field_by_id('nirspty').value
                sptype = sptype.decode('utf-8')

            except KeyError:
                try:
                    sptype = table.get_field_by_id('optspty').value
                    sptype = sptype.decode('utf-8')

                except KeyError:
                    sptype = 'None'

            sptype = util.update_sptype(np.array([sptype]))[0]

            h_flux, _ = h_twomass.magnitude_to_flux(h_mag, None, h_zp)
            phot = h_twomass.spectrum_to_photometry(wavelength, flux) # Normalized units

            flux *= h_flux/phot # [W m-2 micron-1]

            spdata = np.vstack((wavelength, flux))

            simbad_id, distance = queries.get_distance('2MASS '+twomass_id.decode('utf-8')) # [pc]

            dset = database.create_dataset('spectra/spex/'+name, data=spdata)

            dset.attrs['name'] = str(name)
            dset.attrs['sptype'] = str(sptype)
            dset.attrs['simbad'] = str(simbad_id)
            dset.attrs['2MASS/2MASS.J'] = j_mag
            dset.attrs['2MASS/2MASS.H'] = h_mag
            dset.attrs['2MASS/2MASS.Ks'] = ks_mag
            dset.attrs['distance'] = distance # [pc]

    sys.stdout.write('\rAdding SpeX Prism Spectral Library... '+'{:<40}'.format('[DONE]')+'\n')
    sys.stdout.flush()

    database.close()



'''
Read module.
'''

import os
import configparser

import h5py
import numpy as np

from species.analysis import photometry


class ReadObject:
    '''
    Text
    '''

    def __init__(self,
                 object_name):
        '''
        :param object_name: Object name.
        :type object_name: str

        :return: None
        '''

        self.object_name = object_name

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def get_photometry(self,
                       filter_name):
        '''
        :param filter_name: Filter name.
        :type filter_name: str

        :return: Apparent magnitude (mag), magnitude error (error),
                 apparent flux (W m-2 micron-1), flux error (W m-2 micron-1).
        :rtype: numpy.ndarray
        '''

        h5_file = h5py.File(self.database, 'r')
        obj_phot = np.asarray(h5_file['objects/'+self.object_name+'/'+filter_name])
        h5_file.close()

        return obj_phot

    def get_distance(self):
        '''
        :return: Distance (pc).
        :rtype: float
        '''

        h5_file = h5py.File(self.database, 'r')
        obj_distance = np.asarray(h5_file['objects/'+self.object_name+'/distance'])
        h5_file.close()

        return float(obj_distance)

    def get_absmag(self,
                   filter_name):
        '''
        :param filter_name: Filter name.
        :type filter_name: str

        :return: Absolute magnitude (mag), magnitude error (error).
        :rtype: float, float
        '''

        h5_file = h5py.File(self.database, 'r')
        obj_distance = np.asarray(h5_file['objects/'+self.object_name+'/distance'])
        obj_phot = np.asarray(h5_file['objects/'+self.object_name+'/'+filter_name])
        h5_file.close()

        abs_mag = photometry.apparent_to_absolute(obj_phot[0], obj_distance)

        return abs_mag, obj_phot[1]



#!/usr/bin/env python

from setuptools import setup

try:
    from pip._internal.req import parse_requirements
except ImportError:
    from pip.req import parse_requirements

reqs = parse_requirements('requirements.txt', session='hack')
reqs = [str(ir.req) for ir in reqs]

setup(
    name='species',
    version='0.0.6',
    description='Toolkit for analyzing spectral and photometric data of planetary and substellar atmospheres',
    long_description=open('README.rst').read(),
    author='Tomas Stolker',
    author_email='tomas.stolker@phys.ethz.ch',
    url='https://github.com/tomasstolker/species',
    packages=['species',
              'species.analysis',
              'species.core',
              'species.data',
              'species.plot',
              'species.read'],
    package_dir={'species':'species'},
    include_package_data=True,
    install_requires=reqs,
    license='GPLv3',
    zip_safe=False,
    keywords='species',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import os
import shutil

import numpy as np

import species
from . import tools


class TestPhotometry:

    def setup_class(self):
        self.limit = 1e-10

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        os.remove('photometry.pdf')
        shutil.rmtree("data/")

    def test_species_init(self):
        tools.create_config()
        species.SpeciesInit("./")

    def test_spectral_library(self):
        spectrum = species.ReadSpectrum(spectrum='irtf', filter_name='MKO/NSFCam.H')
        specbox = spectrum.get_spectrum(sptype='L0')

        wavelength = specbox.wavelength[0]
        flux = specbox.flux[0]

        assert wavelength.shape == (1063, )
        assert flux.shape == (1063, )

        assert np.allclose(np.sum(wavelength), 1692.8604, rtol=1e-7, atol=0.)
        assert np.allclose(np.sum(flux), 4.5681937e-11, rtol=1e-8, atol=0.)

        spectrum = species.ReadSpectrum(spectrum='irtf', filter_name='MKO/NSFCam.H')
        specbox = spectrum.get_spectrum(sptype='L0')

        synphot = species.SyntheticPhotometry(filter_name='MKO/NSFCam.H')
        phot = synphot.spectrum_to_photometry(wavelength=wavelength,
                                              flux_density=flux)

        assert np.allclose(phot, 4.567475834590298e-14, rtol=self.limit, atol=0.)

        transmission = species.ReadFilter(filter_name='MKO/NSFCam.H')
        wl_mean = transmission.mean_wavelength()

        assert np.allclose(wl_mean, 1.6298258, rtol=1e-7, atol=0.)

        photbox = species.create_box(boxtype='photometry',
                                     name='L0 dwarf',
                                     wavelength=wl_mean,
                                     flux=phot)

        assert photbox.name == 'L0 dwarf'
        assert np.allclose(photbox.wavelength, 1.6298258, rtol=1e-7, atol=0.)
        assert np.allclose(photbox.flux, 4.567475834590298e-14, rtol=self.limit, atol=0.)

        species.plot_spectrum(boxes=(specbox, photbox),
                              filters=('MKO/NSFCam.H', ),
                              output='photometry.pdf',
                              xlim=(1., 2.5),
                              offset=(-0.08, -0.06))

        assert os.path.exists('photometry.pdf')



'''
Photometry module.
'''

import os
import configparser

import h5py
import numpy as np

from scipy.integrate import simps

from species.data import database
from species.core import box
from species.read import read_filter, read_calibration, read_model


def multi_photometry(datatype,
                     spectrum,
                     filters,
                     parameters):
    '''
    :param datatype: Data type ('model' or 'calibration').
    :type datatype: str
    :param spectrum: Spectrum name (e.g., 'drift-phoenix').
    :type spectrum: str
    :param filters: Filter IDs.
    :type filters: tuple(str, )
    :param parameters: Model parameter values.
    :type parameters: dict

    :return: Box with synthetic photometry.
    :rtype: species.core.box.SynphotBox
    '''

    flux = {}

    if datatype == 'model':
        for item in filters:
            readmodel = read_model.ReadModel(spectrum, item)
            flux[item] = readmodel.get_photometry(parameters, ('specres', 100.))

    elif datatype == 'calibration':
        for item in filters:
            readcalib = read_calibration.ReadCalibration(spectrum, item)
            flux[item] = readcalib.get_photometry(parameters)

    return box.create_box('synphot', name='synphot', flux=flux)


def apparent_to_absolute(app_mag,
                         distance):
    '''
    :param app_mag: Apparent magnitude (mag).
    :type app_mag: float or numpy.ndarray
    :param distance: Distance (pc).
    :type distance: float or numpy.ndarray

    :return: Absolute magnitude (mag).
    :rtype: float or numpy.ndarray
    '''

    return app_mag - 5.*np.log10(distance) + 5.


class SyntheticPhotometry:
    '''
    Text
    '''

    def __init__(self,
                 filter_name):
        '''
        :param filter_name: Filter name.
        :type filter_name: str

        :return: None
        '''

        self.filter_name = filter_name
        self.filter_interp = None

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def zero_point(self,
                   wl_range):
        '''
        :param wl_range: Wavelength range (micron). The range from the filter transmission
                         curve is used if set to None.
        :type wl_range: float

        :return: tuple(float, float)
        :rtype:
        '''

        if wl_range is None:
            transmission = read_filter.ReadFilter(self.filter_name)
            wl_range = transmission.wavelength_range()

        h5_file = h5py.File(self.database, 'r')

        try:
            h5_file['spectra/calibration/vega']

        except KeyError:
            h5_file.close()
            species_db = database.Database()
            species_db.add_spectrum('vega')
            h5_file = h5py.File(self.database, 'r')

        readcalib = read_calibration.ReadCalibration('vega', None)
        calibbox = readcalib.get_spectrum()

        wavelength = calibbox.wavelength
        flux = calibbox.flux

        wavelength_crop = wavelength[(wavelength > wl_range[0]) & (wavelength < wl_range[1])]
        flux_crop = flux[(wavelength > wl_range[0]) & (wavelength < wl_range[1])]

        h5_file.close()

        return self.spectrum_to_photometry(wavelength_crop, flux_crop)

    def spectrum_to_photometry(self,
                               wavelength,
                               flux_density):
        '''
        :param wavelength: Wavelength (micron).
        :type wavelength: numpy.ndarray
        :param flux_density: Flux density (W m-2 micron-1).
        :type flux_density: numpy.ndarray

        :return: Average flux density (W m-2 micron-1).
        :rtype: float or numpy.ndarray
        '''

        if self.filter_interp is None:
            transmission = read_filter.ReadFilter(self.filter_name)
            self.filter_interp = transmission.interpolate()

        if isinstance(wavelength[0], (np.float32, np.float64)):
            transmission = self.filter_interp(wavelength)

            indices = np.isnan(transmission)
            indices = np.logical_not(indices)

            integrand1 = transmission[indices]*flux_density[indices]
            integrand2 = transmission[indices]

            integral1 = simps(integrand1, wavelength[indices])
            integral2 = simps(integrand2, wavelength[indices])

            photometry = integral1/integral2

        else:
            photometry = []
            for i, _ in enumerate(wavelength):
                transmission = self.filter_interp(wavelength[i])

                indices = np.isnan(transmission)
                indices = np.logical_not(indices)

                integrand1 = transmission[indices]*flux_density[i][indices]
                integrand2 = transmission[indices]

                integral1 = simps(integrand1, wavelength[i][indices])
                integral2 = simps(integrand2, wavelength[i][indices])

                photometry.append(integral1/integral2)

            photometry = np.asarray(photometry)

        return photometry

    def spectrum_to_magnitude(self,
                              wavelength,
                              flux_density,
                              distance=None):
        '''
        :param wavelength: Wavelength (micron).
        :type wavelength: numpy.ndarray
        :param flux_density: Flux density (W m-2 micron-1).
        :type flux_density: numpy.ndarray
        :param distance: Distance (pc). No absolute magnitude is calculated if set to None.
        :type distance: float

        :return: Flux (W m-2).
        :rtype: float or numpy.ndarray
        '''

        vega_mag = 0.03 # [mag]

        flux = self.spectrum_to_photometry(wavelength, flux_density)
        zp_flux = self.zero_point((wavelength[0], wavelength[-1]))

        # indices = np.isnan(distance)
        # indices = np.logical_not(indices)
        # indices = np.where(indices)[0]
        #
        # flux = flux[indices]
        # distance = distance[indices]

        app_mag = vega_mag - 2.5*np.log10(flux/zp_flux)

        if distance is None:
            abs_mag = None
        else:
            abs_mag = app_mag - 5.*np.log10(distance) + 5.

        return app_mag, abs_mag

    def magnitude_to_flux(self,
                          magnitude,
                          error,
                          zp_flux=None):
        '''
        :param magnitude: Magnitude (mag).
        :type magnitude: float
        :param error: Error (mag). Not used if set to None.
        :type error: float
        :param zp_flux: Zero point flux (W m-2 micron-1). Calculated if set to None.
        :type zp_flux: float

        :return: Flux (W m-2 micron-1), lower error, upper error
        :rtype: float, tuple(float, float)
        '''

        vega_mag = 0.03 # [mag]

        if zp_flux is None:
            zp_flux = self.zero_point(None)

        flux = 10.**(-0.4*(magnitude-vega_mag))*zp_flux

        if error is None:
            error_flux = None

        else:
            error_upper = flux * (10.**(0.4*error) - 1.)
            error_lower = flux * (1. - 10.**(-0.4*error))
            error_flux = (error_lower+error_upper)/2.

        return flux, error_flux

    def flux_to_magnitude(self,
                          flux,
                          distance):
        '''
        :param flux: Flux density (W m-2 micron-1).
        :type flux: float
        :param error: Distance (pc).
        :type error: float

        :return: Apparent magnitude (mag), absolute magnitude (mag).
        :rtype: float, float
        '''

        vega_mag = 0.03 # [mag]

        zp_flux = self.zero_point(None)

        app_mag = vega_mag - 2.5*np.log10(flux/zp_flux)

        if distance is None:
            abs_mag = None
        else:
            abs_mag = app_mag - 5.*np.log10(distance) + 5.

        return app_mag, abs_mag



'''
Utility functions.
'''

import numpy as np


def update_sptype(sptypes):
    '''
    Function to update a list with spectral types to two characters (e.g., M8, L3, or T1).

    :param sptypes: Spectral types.
    :type sptypes: numpy.ndarray

    :return: Updated spectral types.
    :rtype: numpy.ndarray
    '''

    mlty = ('M', 'L', 'T', 'Y')

    for i, spt in enumerate(sptypes):
        if spt == 'None':
            pass

        elif spt == 'null':
            sptypes[i] = 'None'

        else:
            for item in mlty:
                try:
                    sp_index = spt.index(item)
                    sptypes[i] = spt[sp_index:sp_index+2]

                except ValueError:
                    pass

    return sptypes


def update_filter(filter_in):
    '''
    Function to update filter ID from the Vizier Photometry viewer VOTable to the filter ID from
    the SVO Filter Profile Service.

    :param filter_in: Filter ID in the format of the Vizier Photometry viewer
    :type filter_in: str

    :return: Filter ID in the format of the SVO Filter Profile Service.
    :rtype: str
    '''

    if filter_in[0:5] == b'2MASS':
        filter_out = str(b'2MASS/2MASS.'+filter_in[6:])

    elif filter_in[0:4] == b'WISE':
        filter_out = str(b'WISE/WISE.'+filter_in[5:])

    elif filter_in[0:10] == b'GAIA/GAIA2':
        filter_out = str(filter_in[0:9]+b'0'+filter_in[10:])

    else:
        filter_out = None

    return filter_out


def sort_data(teff,
              logg,
              feh,
              wavelength,
              flux):
    '''
    :param teff:
    :type teff: numpy.ndarray
    :param logg:
    :type logg: numpy.ndarray
    :param feh:
    :type feh: numpy.ndarray
    :param wavelength:
    :type wavelength: numpy.ndarray
    :param flux:
    :type flux: numpy.ndarray

    :return:
    :rtype: tuple(numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)
    '''

    teff_unique = np.unique(teff)
    logg_unique = np.unique(logg)
    feh_unique = np.unique(feh)

    spectrum = np.zeros((teff_unique.shape[0],
                         logg_unique.shape[0],
                         feh_unique.shape[0],
                         wavelength.shape[0]))

    for i in range(teff.shape[0]):
        index_teff = np.argwhere(teff_unique == teff[i])[0]
        index_logg = np.argwhere(logg_unique == logg[i])[0]
        index_feh = np.argwhere(feh_unique == feh[i])[0]

        spectrum[index_teff, index_logg, index_feh, :] = flux[i]

    return (teff_unique, logg_unique, feh_unique, wavelength, spectrum)


def write_data(model,
               database,
               data_sorted):
    '''
    :param model: Atmosphere model.
    :type model: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    if 'models/'+model in database:
        del database['models/'+model]

    dset = database.create_group('models/'+model)

    dset.attrs['nparam'] = int(3)
    dset.attrs['parameter0'] = str('teff')
    dset.attrs['parameter1'] = str('logg')
    dset.attrs['parameter2'] = str('feh')

    database.create_dataset('models/'+model+'/teff',
                            data=data_sorted[0],
                            dtype='f')

    database.create_dataset('models/'+model+'/logg',
                            data=data_sorted[1],
                            dtype='f')

    database.create_dataset('models/'+model+'/feh',
                            data=data_sorted[2],
                            dtype='f')

    database.create_dataset('models/'+model+'/wavelength',
                            data=data_sorted[3],
                            dtype='f')

    database.create_dataset('models/'+model+'/flux',
                            data=data_sorted[4],
                            dtype='f')


def add_missing(model,
                database):
    '''
    :param model: Atmosphere model.
    :type model: str
    :param database:
    :type database: h5py._hl.files.File

    :return: None
    '''

    teff = np.asarray(database['models/'+model+'/teff'])
    logg = np.asarray(database['models/'+model+'/logg'])
    feh = np.asarray(database['models/'+model+'/feh'])
    flux = np.asarray(database['models/'+model+'/flux'])

    for i in range(teff.shape[0]):
        for j in range(logg.shape[0]):
            for k in range(feh.shape[0]):
                if np.count_nonzero(flux[i, j, k]) == 0:
                    scaling = (teff[i+1]-teff[i])/(teff[i+1]-teff[i-1])
                    flux[i, j, k] = scaling*flux[i+1, j, k] + (1.-scaling)*flux[i-1, j, k]

    del database['models/'+model+'/flux']

    database.create_dataset('models/'+model+'/flux',
                            data=flux,
                            dtype='f')
